import datetime
import json
from typing import Dict, List, Optional

import json_merge_patch
from loguru import logger
from sqlalchemy import Column, DateTime, MetaData, Table, func, text
from sqlalchemy.dialects import postgresql
from sqlalchemy.engine import Row
from sqlalchemy.exc import (
    IntegrityError,
    NoResultFound,
    OperationalError,
    ProgrammingError,
)
from sqlalchemy.future import Engine, create_engine

from contaxy.operations import JsonDocumentOperations
from contaxy.schema.exceptions import (
    ClientValueError,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
    ServerBaseError,
)
from contaxy.schema.json_db import JsonDocument
from contaxy.utils.postgres_utils import create_schema
from contaxy.utils.state_utils import GlobalState, RequestState


class PostgresJsonDocumentManager(JsonDocumentOperations):
    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
    ):
        """Initializes the Postgres Json Document Manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
        """
        self.global_state = global_state
        self.request_state = request_state
        self._engine = self._create_db_engine()
        self._metadata = MetaData()

    def create_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
        upsert: bool = True,
    ) -> JsonDocument:
        """Creates a json document for a given key.

        An upsert strategy is used, i.e. if a document already exists for the given key it will be overwritten. The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            key (str): Json Document Id, i.e. DB row key.
            json_document (Dict): The actual Json document.
            upsert (bool): Indicates, wheter upsert strategy is used.

        Raises:
            ClientValueError: If the given json_document does not contain valid json.
            ResourceAlreadyExistsError: If a document already exists for the given key and `upsert` is False.

        Returns:
            JsonDocument: The created Json document.
        """
        try:
            json_dict = json.loads(json_document)
        except json.decoder.JSONDecodeError:
            raise ClientValueError("Invalid Json provided")

        table = self._get_collection_table(project_id, collection_id)

        insert_data = {"key": key, "json_value": json_dict}
        insert_data = self._add_metadata_for_insert(insert_data)
        upsert_data = self._add_metadata_for_update(insert_data)

        stmt = postgresql.insert(table).values(**insert_data)
        if upsert:
            stmt = stmt.on_conflict_do_update(index_elements=["key"], set_=upsert_data)

        with self._engine.begin() as conn:
            try:
                result = conn.execute(stmt)
                if result.rowcount == 0:
                    raise ServerBaseError(
                        f"Json Document creation for key {key} for an unknown reason"
                    )
                conn.commit()
            except IntegrityError:
                raise ResourceAlreadyExistsError(
                    f"A Json document for key {key} already exists."
                )

        return self.get_json_document(project_id, collection_id, key)

    def get_json_document(
        self, project_id: str, collection_id: str, key: str
    ) -> JsonDocument:
        """Get a Json document by key.

        The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            key (str): Json Document Id, i.e. DB row key.
            json_document (Dict): The actual Json document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.

        Returns:
            JsonDocument: The requested Json document.
        """
        table = self._get_collection_table(project_id, collection_id)
        select_statement = table.select().where(table.c.key == key)
        with self._engine.begin() as conn:
            result = conn.execute(select_statement)
            try:
                row = result.one()
            except NoResultFound:
                raise ResourceNotFoundError(f"No document with key {key} found")
        return self._map_db_row_to_document_model(row)

    def update_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
    ) -> JsonDocument:
        """Updates a Json document via Json Merge Patch strategy.

        The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            key (str): Json Document Id, i.e. DB row key.
            json_document (Dict): The actual Json document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.
            ServerBaseError: Document not updatded for an unknown reason.

        Returns:
            JsonDocument: The updated document.
        """
        table = self._get_collection_table(project_id, collection_id)

        update_data = self._add_metadata_for_update({})

        select_statement = table.select().with_for_update().where(table.c.key == key)

        with self._engine.begin() as conn:

            result = conn.execute(select_statement)

            if result.rowcount == 0:
                raise ResourceNotFoundError(
                    f"Update failed - No document with key {key} found"
                )
            row = result.one()

            update_data["json_value"] = json_merge_patch.merge(
                row["json_value"], json.loads(json_document)
            )

            # The json_value needs to be a dict otherwise the string gets escaped
            update_statement = (
                table.update().where(table.c.key == key).values(**update_data)
            )

            result = conn.execute(update_statement)

            conn.commit()

        return self.get_json_document(project_id, collection_id, key)

    def delete_json_document(
        self, project_id: str, collection_id: str, key: str
    ) -> None:
        """Delete a Json document by key.

        The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            key (str): Json Document Id, i.e. DB row key.
            json_document (Dict): The actual Json document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.
            ServerBaseError: Document not deleted for an unknown reason.
        """

        table = self._get_collection_table(project_id, collection_id)
        delete_statement = table.delete().where(table.c.key == key)
        with self._engine.begin() as conn:
            result = conn.execute(delete_statement)
            if result.rowcount == 0:
                # This will raise a ResourceNotFoundError if doc not exists
                self.get_json_document(project_id, collection_id, key)
                raise ServerBaseError(
                    f"Document {key} could not be deleted (project_id: {project_id}, collection_id {collection_id})"
                )
            conn.commit()

    def delete_documents(
        self, project_id: str, collection_id: str, keys: List[str]
    ) -> int:
        """Delete Json documents by key.

        The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            keys (List[str]): Json Document Ids, i.e. DB row keys.
            json_document (Dict): The actual Json document.

        """
        table = self._get_collection_table(project_id, collection_id)
        delete_statement = table.delete().where(table.c.key.in_(keys))
        with self._engine.begin() as conn:
            result = conn.execute(delete_statement)
            conn.commit()

        return result.rowcount

    def list_json_documents(
        self,
        project_id: str,
        collection_id: str,
        filter: Optional[str] = None,
        keys: Optional[List[str]] = None,
    ) -> List[JsonDocument]:
        """List all existing Json documents and optionally filter via Json Path syntax.

            The project is equivalent to the DB schema and the collection to a DB table inside the respective DB schema. Schema as well as table will be lazily created.

        Args:
            project_id (str): Project Id, i.e. DB schema.
            collection_id (str): Json document collection Id, i.e. DB table.
            filter (Optional[str], optional): Json Path filter. Defaults to None.
            keys (Optional[List[str]], optional): Json Document Ids, i.e. DB row keys. Defaults to None.

        Raises:
            ClientValueError: If filter is provided and does not contain a valid Json Path filter.

        Returns:
            List[JsonDocument]: List of Json documents.
        """

        table = self._get_collection_table(project_id, collection_id)
        sql_statement = table.select()
        if filter:
            sql_statement = sql_statement.where(
                func.jsonb_path_exists(table.c.json_value, filter),
            )

        if keys:
            sql_statement = sql_statement.where(table.c.key.in_(keys))

        with self._engine.begin() as conn:
            try:
                result = conn.execute(sql_statement)
            except ProgrammingError:
                raise ClientValueError("Please provide a valid Json Path filter.")

            rows = result.fetchall()
        return self._map_db_rows_to_document_models(rows)

    def delete_json_collections(
        self,
        project_id: str,
    ) -> None:
        """Deletes all JSON collections for a project.

        Args:
            project_id: Project ID associated with the collections.
        """
        # TODO: Check if further error handling is needed
        with self._engine.begin() as conn:
            stmt = text(
                f'DROP SCHEMA IF EXISTS "{self._get_schema_name(project_id)}" cascade'
            )
            conn.execute(stmt)
            conn.commit()

    def delete_json_collection(
        self,
        project_id: str,
        collection_id: str,
    ) -> None:
        """Delete a JSON collection.

        Args:
            project_id (str): Project ID associated with the collection.
            collection_id (str): The collection to be deleted.
        """

        stmt = text(
            f'DROP TABLE IF EXISTS "{self._get_schema_name(project_id)}"."{collection_id}" cascade'
        )
        with self._engine.begin() as conn:
            conn.execute(stmt)
            conn.commit()

    def _add_metadata_for_insert(self, data: dict) -> dict:
        # TODO: Copy required?
        insert_data = data.copy()
        # TODO: Finalize
        insert_data["created_at"] = datetime.datetime.utcnow()
        # data["created_by"] =
        return insert_data

    def _add_metadata_for_update(self, data: dict) -> dict:
        update_data = data.copy()
        # TODO: Finalize
        update_data["updated_at"] = datetime.datetime.utcnow()
        # data["updated_by"] =
        return update_data

    def _map_db_rows_to_document_models(self, rows: List[Row]) -> List[JsonDocument]:
        docs = []
        for row in rows:
            docs.append(self._map_db_row_to_document_model(row))
        return docs

    def _map_db_row_to_document_model(self, row: Row) -> JsonDocument:
        data: Dict = {}
        for column_name, value in row._mapping.items():
            if column_name == "json_value":
                value = json.dumps(value)
            data.update({column_name: value})
        return JsonDocument(**data)

    def _get_collection_table(self, project_id: str, collection_id: str) -> Table:
        # TODO: Decide on actual column datatypes
        collection = Table(
            collection_id,
            self._metadata,
            Column("key", postgresql.VARCHAR, primary_key=True),
            Column("json_value", postgresql.JSONB),
            Column("created_at", DateTime),
            Column("created_by", postgresql.VARCHAR),
            Column("updated_at", DateTime),
            Column("updated_by", postgresql.VARCHAR),
            schema=self._get_schema_name(project_id),
            keep_existing=True,  # TODO: Depends on how we handle schema modifications
        )

        try:
            collection.create(self._engine, checkfirst=True)
        except ProgrammingError:
            create_schema(self._engine, self._get_schema_name(project_id))
            collection.create(self._engine, checkfirst=True)

        return collection

    def _create_db_engine(self) -> Engine:
        state_namespace = self.global_state[PostgresJsonDocumentManager]
        if not state_namespace.db_engine:
            url = self.global_state.settings.POSTGRES_CONNECTION_URI
            engine = create_engine(url, future=True)
            # Test the DB connection and set to global state if succesful
            try:
                with engine.begin():
                    state_namespace.db_engine = engine
            except OperationalError as ex:
                logger.exception("POSTGRES DB Problem")
                raise ServerBaseError(
                    "Postgres DB connection failed. Validate connection URI."
                ) from ex
            logger.info("Postgres DB Engine created")
        return state_namespace.db_engine

    def _get_schema_name(self, project_id: str) -> str:
        prefix = self.global_state.settings.SYSTEM_NAMESPACE
        return project_id if not prefix else f"{prefix}_{project_id}"



from typing import Dict, List, Tuple

from contaxy import config
from contaxy.operations import AuthOperations
from contaxy.operations.project import ProjectOperations
from contaxy.schema import Project, User
from contaxy.schema.auth import USER_ROLE, USERS_KIND, AccessLevel, UserRegistration
from contaxy.schema.exceptions import ClientValueError
from contaxy.schema.project import ProjectCreation

PERMISSION_SEPERATOR = "#"
RESOURCE_WILDCARD = "*"  # Allows access to all resources

ACCESS_LEVEL_MAPPING: Dict[AccessLevel, List[AccessLevel]] = {
    AccessLevel.ADMIN: [AccessLevel.WRITE, AccessLevel.READ],
    AccessLevel.WRITE: [AccessLevel.READ],
    AccessLevel.READ: [],
}

ACCESS_LEVEL_REVERSE_MAPPING: Dict[AccessLevel, List[AccessLevel]] = {}
for key, value in ACCESS_LEVEL_MAPPING.items():
    for string in value:
        ACCESS_LEVEL_REVERSE_MAPPING.setdefault(string, []).append(key)


def is_valid_permission(permission_str: str) -> bool:
    """Returns `True` if the `permission_str` is valid permission."""
    # TODO: do some basic checks on the permission string
    return PERMISSION_SEPERATOR in permission_str


def is_jwt_token(token: str) -> bool:
    """Returns `True` if the provided token is an JWT token."""
    # TODO: Improve this detection
    return len(token) != config.settings.API_TOKEN_LENGTH and not token.islower()


def construct_permission(resource_name: str, access_level: AccessLevel) -> str:
    """Constructs a permission based on the provided `resource_name`  and `access_level`."""
    return resource_name + PERMISSION_SEPERATOR + access_level.value


def parse_permission(permission: str) -> Tuple[str, AccessLevel]:
    """Extracts the resource name and access level from a permission.

    Args:
        permission: The permission to parse.

    Raises:
        ClientValueError: If the permission cannot be parsed.

    Returns:
        Tuple[str, AccessLevel]: The extracted resource name and access level.
    """

    if not is_valid_permission(permission):
        raise ClientValueError(f"{permission} is not a valid permission.")

    resource, level_str = permission.split(PERMISSION_SEPERATOR)
    # The resource name should not start or end with a slash or colon
    resource = resource.lstrip("/").lstrip(":").rstrip("/").rstrip(":")
    return resource, AccessLevel.load(level_str)


def is_access_level_granted(
    granted_access_level: AccessLevel, requested_access_level: AccessLevel
) -> bool:
    """Checks if the requested access level is allowed by the granted access level.

    Args:
        granted_access_level: The access level that is already granted.
        requested_access_level: The access level to check against the granted level.

    Returns:
        bool: `True` if the access level is granted.
    """
    if granted_access_level is requested_access_level:
        # Same access level
        return True

    return requested_access_level in ACCESS_LEVEL_MAPPING[granted_access_level]


def is_permission_granted(granted_permission: str, requested_permission: str) -> bool:
    """Checks if the requested permission is allowed by the granted permission.

    Args:
        granted_permission: The permission that is already granted.
        requested_permission: The permission to check against the granted permission

    Raises:
        ClientValueError: If one of the permissions cannot be parsed.

    Returns:
        bool: `True` if the permission is granted.
    """

    granted_perm_resource, granted_perm_level = parse_permission(granted_permission)
    requested_perm_resource, requested_perm_level = parse_permission(
        requested_permission
    )

    # the * is a special -> it allows access to all resources
    # Add path divider (/ and :) since it should only grant access if it matches a full path
    if (
        granted_perm_resource != RESOURCE_WILDCARD
        and not f"{requested_perm_resource}/".startswith(f"{granted_perm_resource}/")
        and not f"{requested_perm_resource}:".startswith(f"{granted_perm_resource}:")
    ):
        # The requested permission does not start with the granted permission
        return False

    if not is_access_level_granted(granted_perm_level, requested_perm_level):
        # The requested permission needs atleast the same level as the granted permission
        return False

    return True


def create_and_setup_user(
    user_input: UserRegistration,
    auth_manager: AuthOperations,
    project_manager: ProjectOperations,
    technical_user: bool = False,
) -> User:
    """Create a new user and setup default project and permissions.

    Args:
        user_input (UserRegistration): Information required for creating a new user.
        auth_manager (AuthOperations): The auth manager used to setup default permissions.
        project_manager (ProjectOperations): The project manager used to create the default user project.

    Raises:
        ResourceAlreadyExistsError: If the user already exists

    Returns:
        User: The newly created and setup user
    """
    user = auth_manager.create_user(user_input, technical_user)
    _setup_user_default_permissions(user, auth_manager)
    _setup_user_home_project(user, project_manager)
    return user


def _setup_user_default_permissions(user: User, auth_manager: AuthOperations) -> None:
    """Assign the permissions/roles that every user should have by default.

    Args:
        user (User): The user for whom the home project shall be created.
        auth_manager (AuthOperations): The auth manager used to setup the permissions/roles

    Raises:
        ResourceAlreadyExistsError: if a project with the id of the user already exists.

    Returns:
        Project: The created technical project of which the user is a member.
    """
    # TODO: get resource name from object
    user_resource_name = USERS_KIND + "/" + user.id
    # Give user admin permission to itself
    auth_manager.add_permission(
        user_resource_name,
        construct_permission(user_resource_name, AccessLevel.ADMIN),
    )
    # Add user to the default user role
    auth_manager.add_permission(user_resource_name, USER_ROLE)


def _setup_user_home_project(user: User, project_manager: ProjectOperations) -> Project:
    """Create the user's default home project that he has full access to.

    Args:
        user (User): The user for whom the home project shall be created.
        project_manager (ProjectOperations): The project manager used to create the project.

    Raises:
        ResourceAlreadyExistsError: if a project with the id of the user already exists.

    Returns:
        Project: The created technical project of which the user is a member.
    """
    if user.username is not None:
        # Split name by space and dot
        user_first_name = user.username.split()[0].split(".")[0]
        user_project_name = f"{user_first_name.capitalize()}'s Home"
    else:
        user_project_name = "Home"
    user_project = project_manager.create_project(
        project_input=ProjectCreation(
            id=user.id,
            display_name=user_project_name,
            description="My personal project nobody else has access to.",
            # TODO: Maybe state that only the admin has access to the project
        ),
        technical_project=True,
    )

    if user_project.id:
        project_manager.add_project_member(user_project.id, user.id, AccessLevel.ADMIN)

    return user_project


def parse_userid_from_resource_name(user_resource_name: str) -> str:
    """Returns the user id from a user-resource name.

    For example, 'users/abcd' returns 'abcd' as 'users' is the prefix.

    Args:
        user_resource_name (str): The resource name, e.g. 'users/abcd'.

    Returns:
        str: The user id in the user-resource name or an empty string.
    """

    if not user_resource_name or not user_resource_name.startswith("users/"):
        return ""

    return user_resource_name.split("/")[1]



"""Utilities for generating IDs, tokens, and hashes."""

import hashlib
import math
import re
import secrets
import string
from email.utils import parseaddr
from typing import List, Optional, Tuple

import shortuuid
from slugify import slugify

from contaxy.config import settings

_PROJECT_ID_SEPERATOR = "-p-"
_PROJECT_NAME_TO_ID_REGEX = re.compile(r"^projects/([^/:\s]+)$")
_USER_NAME_TO_ID_REGEX = re.compile(r"^users/([^/:\s]+)$")
_SERVICE_RESOURCE_REGEX = re.compile(r"^projects/([^/:\s]+)/services/([^/:\s]+)")


def is_email(text: str) -> bool:
    """Returns `True` if the given `text` has the format of an email address."""
    if not re.match(r"[^@]+@[^@]+\.[^@]+", text):
        return False
    return "@" in parseaddr(text)[1]


def extract_user_id_from_resource_name(user_resource_name: str) -> str:
    """Extract the user id from a provided resource name.

    Raises:
        ValueError: If the `user_resource_name` is not valid.
    """
    match = re.match(_USER_NAME_TO_ID_REGEX, user_resource_name)
    if match:
        return match[1]
    else:
        raise ValueError(
            f"The provided user_resource_name is not valid: {user_resource_name}"
        )


def extract_project_id_from_resource_name(project_resource_name: str) -> str:
    """Extract the project id from a provided resource name.

    Raises:
        ValueError: If the `project_resource_name` is not valid.
    """
    match = re.match(_PROJECT_NAME_TO_ID_REGEX, project_resource_name)
    if match:
        return match[1]
    else:
        raise ValueError(
            f"The provided project_resource_name is not valid: {project_resource_name}"
        )


def extract_ids_from_service_resource_name(
    service_resource_name: str,
) -> Tuple[str, str]:
    """Extract the project id and service id from a service resource name.

    Args:
        service_resource_name (str): The service resource name (e.g. /projects/project-id/services/service-id)

    Returns:
        Tuple[str, str]: The project id and service id contained in the resource name
    """
    match = _SERVICE_RESOURCE_REGEX.match(service_resource_name)
    if match is None:
        raise ValueError(
            f"Resource name {service_resource_name} is not a service resource!"
        )
    groups = match.groups()
    return groups[0], groups[1]


def get_project_resource_prefix(project_id: str) -> str:
    """Creates a prefix usable to construct IDs for project resources.

    The resource prefix is based on the system namespace and project ID.
    """
    return settings.SYSTEM_NAMESPACE + _PROJECT_ID_SEPERATOR + project_id


def hash_str(input_str: str, length: Optional[int] = None) -> str:
    """Generates a hash with a variable lenght from an abritary text.

    Args:
        input_str (str): Text to hash.
        length (Optional[int], optional): The length of the generated hash. A shorter hash will result in more possible collusions.
    """
    # Alternatives:
    # hashlib.shake_128(input_str.encode('UTF-8')).hexdigest(length)
    # hashlib.md5, hashlib.sha512
    # Base64 encode? base64.urlsafe_b64encode(hasher.digest()[:10])
    # h.digest().encode('base64')[:6]
    hashed_str = hashlib.sha1(input_str.encode("UTF-8")).hexdigest()

    if length:
        if length > len(hashed_str):
            raise ValueError(
                "The length cannot be set to be more then " + str(len(hashed_str))
            )
        hashed_str = hashed_str[:length]

    return hashed_str


def generate_token(length: int) -> str:
    """Generates a random token with the specified length."""
    # secrets.token_hex returns all bytes as hex (2 bytes per character)
    # Alternative: secrets.token_urlsafe(length) -> base64 encoded token
    return str(secrets.token_hex(math.ceil(length / 2)))[:length]


def generate_short_uuid() -> str:
    """Generates a short - 25 chars - UUID by using all ascii letters and digits."""
    shortuuid.set_alphabet(string.ascii_letters.lower() + string.digits)
    return str(shortuuid.uuid())


def generate_readable_id(
    input_str: str,
    max_length: Optional[int] = None,
    min_length: Optional[int] = None,
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    max_hash_suffix_length: Optional[int] = None,
    separator: str = "-",
    stopwords: Optional[List[str]] = None,
) -> str:
    """Generates a human- and URL-friendly ID from arbritary text.

    Args:
        input_str (str): Text to use for ID generation.
        max_length (Optional[int], optional): Maximal length of the ID. Defaults to `None`.
        min_length (Optional[int], optional): Minimal length of the ID. The generated ID will be filled with a suffix based on its hash. Defaults to `None`.
        suffix (Optional[str], optional): Suffix to add to the generated ID. Defaults to `None`.
        prefix (Optional[str], optional): Prefix to add the the generated ID. Defaults to `None`.
        max_hash_suffix_length (Optional[int], optional): Number of chars to append to the generated ID to allow generating unique IDs with a max lengths. Defaults to `None`.
        separator (str, optional): Seperator to use for replacing whitespaces. Defaults to `-`.
        stopwords (Optional[List[str]], optional): List of stopwords to ignore in the generated ID. Defaults to `None`.

    Raises:
        ValueError: A a function parameter has an inappropriate value.

    Returns:
        str: The generated ID human- and URL-friendly ID.
    """
    if not max_length:
        # Set max_length to the length of the input text if not provided
        max_length = len(input_str)

    if max_length <= 0:
        raise ValueError("The max_length should be bigger than 0.")

    adapted_max_length = max_length

    if suffix:
        adapted_max_length = max_length - len(suffix)

    if prefix:
        adapted_max_length = max_length - len(prefix)

    if adapted_max_length <= 0:
        raise ValueError("The max_length should be more than the suffix and/or prefix.")

    if max_hash_suffix_length:
        if max_length < max_hash_suffix_length:
            raise ValueError(
                "The max_length should not be smaller than the hash suffix."
            )
        adapted_max_length = 0

    if min_length and min_length > max_length:
        raise ValueError("The min_length should not be bigger than max_length.")

    if not stopwords:
        stopwords = []

    generated_id = slugify(
        input_str,
        max_length=adapted_max_length,
        stopwords=stopwords,
        separator=separator,
        entities=False,
        decimal=False,
        hexadecimal=False,
        lowercase=True,
        word_boundary=False,
    )

    if max_hash_suffix_length and len(generated_id) > max_length:
        if max_length <= (max_hash_suffix_length + len(separator)):
            # Only return the hash
            return hash_str(generated_id, length=max_hash_suffix_length)[:max_length]

        # shorten
        shortened_id = generated_id[: max_length - max_hash_suffix_length]
        if not shortened_id.endswith(separator):
            # Append minus
            shortened_id = shortened_id[: -len(separator)] + separator

        generated_id = shortened_id + hash_str(
            generated_id, length=max_hash_suffix_length
        )

    if suffix:
        generated_id = generated_id + suffix

    if prefix:
        generated_id = prefix + generated_id

    # Slugify again to include suffix/prefix
    generated_id = slugify(
        generated_id,
        max_length=max_length,
        separator=separator,
        stopwords=stopwords,
        entities=False,
        decimal=False,
        hexadecimal=False,
        lowercase=True,
        word_boundary=False,
    )

    if min_length and len(generated_id) < min_length:
        generated_id += hash_str(generated_id, length=min_length - len(generated_id))

    return generated_id



from datetime import datetime
from typing import Any, List, Literal, Optional

import docker
import docker.errors
from loguru import logger

from contaxy.managers.deployment.docker_utils import (
    create_container_config,
    delete_container,
    get_project_container,
    get_project_containers,
    handle_network,
    list_deploy_service_actions,
    map_job,
    map_service,
    read_container_logs,
    reconnect_to_all_networks,
    wait_for_container,
)
from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType, ServiceUpdate
from contaxy.schema.exceptions import ClientBaseError, ClientValueError


class DockerDeploymentPlatform:
    _is_initialized = False

    def __init__(self) -> None:
        """Initializes the docker deployment manager."""

        self.client = docker.from_env()
        # Reconnect the backend to all existing docker networks on startup
        if not DockerDeploymentPlatform._is_initialized:
            reconnect_to_all_networks(self.client)
            DockerDeploymentPlatform._is_initialized = True

    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        try:
            containers = get_project_containers(
                client=self.client,
                project_id=project_id,
                deployment_type=deployment_type,
            )

            return [map_service(container) for container in containers]
        except docker.errors.APIError:
            return []

    def deploy_service(
        self,
        project_id: str,
        service: Service,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Service:
        container_config = create_container_config(
            deployment=service,
            project_id=project_id,
        )
        handle_network(client=self.client, project_id=project_id)

        try:
            container = self.client.containers.run(**container_config)
            if wait:
                container = wait_for_container(container, self.client)
        except docker.errors.APIError as e:
            logger.error(f"Error in deploy service '{service.display_name}': {e}")
            raise ClientValueError(
                f"Could not deploy service '{service.display_name}'."
            )

        return map_service(container)

    def list_deploy_service_actions(
        self, project_id: str, service: ServiceInput
    ) -> List[ResourceAction]:
        return list_deploy_service_actions(project_id=project_id, deploy_input=service)

    def get_service_metadata(self, project_id: str, service_id: str) -> Service:
        container = get_project_container(
            client=self.client, project_id=project_id, deployment_id=service_id
        )
        return map_service(container)

    def update_service(
        self, project_id: str, service_id: str, service: ServiceUpdate
    ) -> Service:
        # Service update is only implemented on DeploymentManagerWithDB wrapper
        raise NotImplementedError()

    def update_service_access(self, project_id: str, service_id: str) -> None:
        # Service update is only implemented on DeploymentManagerWithDB wrapper
        raise NotImplementedError()

    def delete_service(
        self, project_id: str, service_id: str, delete_volumes: bool = False
    ) -> None:
        container = get_project_container(
            self.client, project_id=project_id, deployment_id=service_id
        )
        delete_container(
            client=self.client, container=container, delete_volumes=delete_volumes
        )

    def delete_services(
        self,
        project_id: str,
    ) -> None:
        containers = get_project_containers(client=self.client, project_id=project_id)
        for container in containers:
            container.remove(v=True, force=True)

    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        container = get_project_container(
            self.client, project_id=project_id, deployment_id=service_id
        )

        return read_container_logs(container=container, lines=lines, since=since)

    def list_jobs(self, project_id: str) -> List[Job]:
        containers = get_project_containers(
            self.client, project_id=project_id, deployment_type=DeploymentType.JOB
        )
        return [map_job(container) for container in containers]

    def deploy_job(
        self,
        project_id: str,
        job: Job,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Job:
        container_config = create_container_config(
            deployment=job,
            project_id=project_id,
        )
        handle_network(client=self.client, project_id=project_id)

        try:
            container = self.client.containers.run(**container_config)
            if wait:
                container = wait_for_container(container, self.client)
        except docker.errors.APIError:
            raise ClientBaseError(
                status_code=500, message=f"Could not deploy job '{job.display_name}'."
            )

        response_service = map_job(container)
        return response_service

    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
    ) -> List[ResourceAction]:
        return list_deploy_service_actions(project_id=project_id, deploy_input=job)

    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        container = get_project_container(
            client=self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )

        return map_job(container)

    def delete_job(self, project_id: str, job_id: str) -> None:
        container = get_project_container(
            self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )
        delete_container(client=self.client, container=container)

    def delete_jobs(
        self,
        project_id: str,
    ) -> None:
        containers = get_project_containers(
            self.client, project_id=project_id, deployment_type=DeploymentType.JOB
        )
        for container in containers:
            container.remove(v=True, force=True)

    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        container = get_project_container(
            self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )

        return read_container_logs(container=container, lines=lines, since=since)

    def suggest_service_config(
        self, project_id: str, container_image: str
    ) -> ServiceInput:
        raise NotImplementedError()

    def list_service_actions(
        self, project_id: str, service_id: str
    ) -> List[ResourceAction]:
        raise NotImplementedError()

    def execute_service_action(
        self, project_id: str, service_id: str, action_id: str
    ) -> Any:
        raise NotImplementedError()

    def suggest_job_config(self, project_id: str, container_image: str) -> JobInput:
        raise NotImplementedError()

    def list_job_actions(self, project_id: str, job_id: str) -> List[ResourceAction]:
        raise NotImplementedError()

    def execute_job_action(self, project_id: str, job_id: str, action_id: str) -> Any:
        raise NotImplementedError()



from .auth import AuthOperations
from .deployment import DeploymentOperations, JobOperations, ServiceOperations
from .extension import ExtensionOperations
from .file import FileOperations
from .json_db import JsonDocumentOperations
from .project import ProjectOperations
from .seed import SeedOperations
from .system import SystemOperations



from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, Path, Query, Request, Response, status
from fastapi.responses import StreamingResponse

from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.managers.extension import parse_composite_id
from contaxy.schema import ExtensibleOperations, File, FileInput, ResourceAction
from contaxy.schema.auth import AccessLevel
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    UPDATE_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
    ClientValueError,
)
from contaxy.schema.extension import EXTENSION_ID_PARAM
from contaxy.schema.file import FILE_KEY_PARAM
from contaxy.schema.project import PROJECT_ID_PARAM
from contaxy.schema.shared import OPEN_URL_REDIRECT, RESOURCE_ID_REGEX
from contaxy.utils.file_utils import FormMultipartStream, SyncFromAsyncGenerator

router = APIRouter(
    tags=["files"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)


@router.get(
    "/projects/{project_id}/files",
    operation_id=ExtensibleOperations.LIST_FILES.value,
    response_model=List[File],
    summary="List project files.",
    status_code=status.HTTP_200_OK,
)
def list_files(
    project_id: str = PROJECT_ID_PARAM,
    recursive: bool = Query(True, description="Include all content of subfolders."),
    include_versions: bool = Query(
        False,
        description="Include all versions of all files.",
    ),
    prefix: Optional[str] = Query(
        None,
        description="Filter results to include only files whose names begin with this prefix.",
    ),
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all available files for the project.

    The files can be filtered by using a `prefix`. The prefix is applied on the full path (directory path + filename).

    All versions of the files can be included by setting `versions` to `true` (default is `false`).

    Set `recursive` to `false` to only show files and folders (prefixes) of the current folder.
    The current folder is either the root folder (`/`) or the folder selected by the `prefix` parameter (has to end with `/`).
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/files", AccessLevel.READ
    )

    return component_manager.get_file_manager(extension_id).list_files(
        project_id, recursive, include_versions, prefix
    )


@router.post(
    "/projects/{project_id}/files/{file_key:path}",
    operation_id=ExtensibleOperations.UPLOAD_FILE.value,
    response_model=File,
    summary="Upload a file.",
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def upload_file(
    request: Request,
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Uploads a file to a file storage.

    The file will be streamed to the selected file storage (core platform or extension).

    This upload operation only supports to attach a limited set of file metadata.
    Once the upload is finished, you can use the [update_file_metadata operation](#files/update_file_metadata)
    to add or update the metadata of the files.

    The `file_key` allows to categorize the uploaded file under a virtual file structure managed by the core platform.
    This allows to create a directory-like structure for files from different extensions and file-storage types.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/files", AccessLevel.WRITE
    )

    file_stream = SyncFromAsyncGenerator(
        request.stream(), component_manager.global_state.shared_namespace.async_loop
    )

    multipart_stream = FormMultipartStream(
        file_stream, request.headers, form_field="file", hash_algo="md5"
    )
    content_type = (
        multipart_stream.content_type
        if multipart_stream.content_type
        else "application/octet-stream"
    )
    return component_manager.get_file_manager().upload_file(
        project_id, file_key, multipart_stream, content_type
    )


@router.post(
    "/projects/{project_id}/multipart-upload",
    operation_id=ExtensibleOperations.UPLOAD_FILE_NO_KEY.value,
    response_model=File,
    summary="Upload a file via multipart stream with filename as file key.",
    status_code=status.HTTP_200_OK,
    # TODO: Decide if to include
    include_in_schema=False,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def upload_file_without_key(
    request: Request,
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Uploads a file to a file storage.

    The file will be streamed to the selected file storage (core platform or extension).

    The file key will be derived based on the filename in the multipart stream.

    This upload operation only supports to attach a limited set of file metadata.
    Once the upload is finished, you can use the [update_file_metadata operation](#files/update_file_metadata)
    to add or update the metadata of the files.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/files", AccessLevel.WRITE
    )

    file_stream = SyncFromAsyncGenerator(
        request.stream(), component_manager.global_state.shared_namespace.async_loop
    )

    multipart_stream = FormMultipartStream(
        file_stream, request.headers, form_field="file", hash_algo="md5"
    )

    content_type = (
        multipart_stream.content_type
        if multipart_stream.content_type
        else "application/octet-stream"
    )

    if not multipart_stream.filename:
        raise ClientValueError(
            "The multipart stream does not contain a file name. Use the endpoint `/projects/{project_id}/files/{file_key:path/}` to explicitly provide a file key."
        )

    return component_manager.get_file_manager().upload_file(
        project_id, multipart_stream.filename, multipart_stream, content_type
    )


@router.get(
    "/projects/{project_id}/files/{file_key:path}:metadata",
    operation_id=ExtensibleOperations.GET_FILE_METADATA.value,
    response_model=File,
    summary="Get file metadata.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_file_metadata(
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, the latest version will be used.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns metadata about the specified file."""
    component_manager.verify_access(
        token,
        f"/projects/{project_id}/files/{file_key}:metadata",
        AccessLevel.READ,
    )

    file_key, extension_id = parse_composite_id(file_key)
    return component_manager.get_file_manager(extension_id).get_file_metadata(
        project_id, file_key, version
    )


@router.patch(
    "/projects/{project_id}/files/{file_key:path}",
    operation_id=ExtensibleOperations.UPDATE_FILE_METADATA.value,
    response_model=File,
    summary="Update file metadata.",
    status_code=status.HTTP_200_OK,
    responses={**UPDATE_RESOURCE_RESPONSES},
)
def update_file_metadata(
    file: FileInput,
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, the latest version will be used.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Updates the file metadata.

    This will not change the actual content or the key of the file.

    The update is applied on the existing metadata based on the JSON Merge Patch Standard ([RFC7396](https://tools.ietf.org/html/rfc7396)).
    Thereby, only the specified properties will be updated.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/files/{file_key}",
        AccessLevel.WRITE,
    )

    file_key, extension_id = parse_composite_id(file_key)
    return component_manager.get_file_manager(extension_id).update_file_metadata(
        file, project_id, file_key, version
    )


@router.get(
    "/projects/{project_id}/files/{file_key:path}:download",
    operation_id=ExtensibleOperations.DOWNLOAD_FILE.value,
    # TODO: response_model?
    summary="Download a file.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def download_file(
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, the latest version will be used.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Downloads the selected file.

    If the file storage supports versioning and no `version` is specified, the latest version will be downloaded.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/files/{file_key}:download",
        AccessLevel.READ,
    )
    file_manager = component_manager.get_file_manager()
    metadata = file_manager.get_file_metadata(project_id, file_key, version)
    file_stream = file_manager.download_file(project_id, file_key, version)

    return StreamingResponse(
        file_stream,
        media_type=metadata.content_type,
        headers={"Content-Disposition": f"attachment;filename={file_key}"},
    )


@router.get(
    "/projects/{project_id}/files/{file_key:path}/actions",
    operation_id=ExtensibleOperations.LIST_FILE_ACTIONS.value,
    response_model=List[ResourceAction],
    summary="List file actions.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def list_file_actions(
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, the latest version will be used.",
    ),
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all actions available for the specified file.

    The returned action IDs should be used when calling the [execute_file_action](#files/execute_file_action) operation.

    The action mechanism allows extensions to provide additional functionality on files. It works the following way:

    1. The user requests all available actions via the [list_file_actions](#files/list_file_actions) operation.
    2. The operation will be forwarded to all installed extensions that have implemented the [list_file_actions](#files/list_file_actions) operation.
    3. Extensions can run arbitrary code - e.g., request and check the file metadata for compatibility - and return a list of actions with self-defined action IDs.
    4. The user selects one of those actions and triggers the [execute_file_action](#files/execute_file_action) operation by providing the selected action- and extension-ID.
    5. The operation is forwarded to the selected extension, which can run arbitrary code to execute the selected action.
    6. The return value of the operation can be either a simple message  (shown to the user) or a redirect to another URL (e.g., to show a web UI).

    The same action mechanism is also used for other resources (e.g., services, jobs).
    It can support a very broad set of use-cases, for example: CSV Viewer, Tensorflow Model Deployment, ZIP Archive Explorer ...
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/files/{file_key}/actions",
        AccessLevel.WRITE,
    )

    return component_manager.get_file_manager(extension_id).list_file_actions(
        project_id, file_key, version
    )


@router.get(
    "/projects/{project_id}/files/{file_key:path}/actions/{action_id}",
    operation_id=ExtensibleOperations.EXECUTE_FILE_ACTION.value,
    # TODO: what is the response model? add additional status codes?
    summary="Execute a file action.",
    status_code=status.HTTP_200_OK,
    responses={**OPEN_URL_REDIRECT, **GET_RESOURCE_RESPONSES},
)
def execute_file_action(
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    action_id: str = Path(
        ...,
        description="The action ID from the file actions operation.",
        regex=RESOURCE_ID_REGEX,
    ),
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, the latest version will be used.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Executes the selected action.

    The actions need to be first requested from the [list_file_actions](#files/list_file_actions) operation.
    If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

    The action mechanism is further explained in the description of the [list_file_actions](#files/list_file_actions).
    """
    # TODO: write permission?
    component_manager.verify_access(
        token,
        f"projects/{project_id}/files/{file_key}/actions/{action_id}",
        AccessLevel.WRITE,
    )

    # TODO: only extract extension ID from action?
    action_id, extension_id = parse_composite_id(action_id)

    return component_manager.get_file_manager(extension_id).execute_file_action(
        project_id, file_key, action_id, version
    )


@router.delete(
    "/projects/{project_id}/files/{file_key:path}",
    operation_id=ExtensibleOperations.DELETE_FILE.value,
    summary="Delete a file.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_file(
    project_id: str = PROJECT_ID_PARAM,
    file_key: str = FILE_KEY_PARAM,
    version: Optional[str] = Query(
        None,
        description="File version tag. If not specified, all versions of the file will be deleted.",
    ),
    keep_latest_version: Optional[bool] = Query(
        False, description="Keep the latest version of the file."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes the specified file.

    If the file storage supports versioning and no `version` is specified, all versions of the file will be deleted.

    The parameter `keep_latest_version` is useful if you want to delete all older versions of a file.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/files/{file_key}", AccessLevel.WRITE
    )
    if keep_latest_version is None:
        keep_latest_version = False

    file_key, extension_id = parse_composite_id(file_key)
    component_manager.get_file_manager(extension_id).delete_file(
        project_id, file_key, version, keep_latest_version
    )
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/projects/{project_id}/files",
    operation_id=ExtensibleOperations.DELETE_FILES.value,
    summary="Delete all files.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_files(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes all files associated with a project."""
    component_manager.verify_access(
        token, f"projects/{project_id}/files", AccessLevel.ADMIN
    )

    component_manager.get_file_manager(extension_id).delete_files(project_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


def modify_openapi_schema(openapi_schema: dict) -> Dict[str, Any]:

    request_body = {
        "required": True,
        "content": {
            "multipart/form-data": {
                "schema": {
                    "$ref": "#/components/schemas/Body_upload_file_projects__project_id__files__file_key___post"
                }
            }
        },
    }

    openapi_schema["paths"]["/projects/{project_id}/files/{file_key}"]["post"][
        "requestBody"
    ] = request_body

    component_schema = {
        "title": "Body_upload_file_projects__project_id__files__file_key___post",
        "required": ["file"],
        "type": "object",
        "properties": {
            "file": {"title": "File", "type": "string", "format": "binary"},
            # "file_name": {"title": "File Name", "type": "string"},
        },
    }

    if (
        "components" not in openapi_schema
        or "schemas" not in openapi_schema["components"]
    ):
        openapi_schema["components"] = {"schemas": {}}

    openapi_schema["components"]["schemas"][
        "Body_upload_file_projects__project_id__files__file_key___post"
    ] = component_schema

    return openapi_schema



from typing import Optional

from fastapi import FastAPI, Request
from loguru import logger
from pydantic.networks import PostgresDsn
from starlette.datastructures import State

from contaxy import config
from contaxy.managers.auth import AuthManager
from contaxy.managers.deployment.manager import DeploymentManager
from contaxy.managers.extension import ExtensionManager
from contaxy.managers.project import ProjectManager
from contaxy.managers.seed import SeedManager
from contaxy.managers.system import SystemManager
from contaxy.operations import (
    FileOperations,
    JobOperations,
    JsonDocumentOperations,
    SeedOperations,
    ServiceOperations,
)
from contaxy.operations.components import ComponentOperations
from contaxy.operations.deployment import DeploymentOperations
from contaxy.schema.auth import AccessLevel, AuthorizedAccess
from contaxy.schema.deployment import ServiceInput
from contaxy.schema.extension import CORE_EXTENSION_ID
from contaxy.utils import auth_utils
from contaxy.utils.state_utils import GlobalState, RequestState


class ComponentManager(ComponentOperations):
    """Initializes and manages all platform components.

    The component manager is created for every request
    and will initialize all platform components based on the platform settings.
    It is the central point to access any other platform component.

    Individual components can store a global state via the `global_state` variable.
    This allows initializing certain objects (DB connections, HTTP clients, ...)
    only once per app instance (process) and share it with other components.

    There is also a `request_state` that can be used to share objects
    for a single request.
    """

    @classmethod
    def from_request(cls, request: Request) -> "ComponentManager":
        return cls(
            GlobalState(request.app.state),
            RequestState(request.state),
        )

    @classmethod
    def from_app(cls, app: FastAPI) -> "ComponentManager":
        return cls(
            GlobalState(app.state),
            RequestState(State()),
        )

    def __init__(self, global_state: GlobalState, request_state: RequestState):
        """Initializes the component manager.

        Args:
            global_state: Global application state.
            request_state: Request scoped state.
        """

        # Individual components can store global state via the `global_state` variable
        self._global_state = global_state
        self._request_state = request_state

        # Initialized variables which will be lazyloaded
        self._auth_manager: Optional[AuthManager] = None
        self._extension_manager: Optional[ExtensionManager] = None
        self._project_manager: Optional[ProjectManager] = None
        self._system_manager: Optional[SystemManager] = None
        # Extensible managers: typed by its interface
        self._json_db_manager: Optional[JsonDocumentOperations] = None
        self._deployment_manager: Optional[DeploymentOperations] = None
        self._file_manager: Optional[FileOperations] = None

        # Debug Manager
        self._seed_manager: Optional[SeedOperations] = None

    # Allow component manger to be used as context manager
    def __enter__(self) -> "ComponentManager":
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        self.close()

    def close(self) -> None:
        """Closes the component manager.

        This is called once the request is finished
        and will close the `request_state` and all its registered close callbacks.
        """
        self.request_state.close()
        del self._request_state

    @property
    def global_state(self) -> GlobalState:
        return self._global_state

    @property
    def request_state(self) -> RequestState:
        return self._request_state

    def verify_access(
        self,
        token: str,
        resource_name: Optional[str] = None,
        access_level: Optional[AccessLevel] = None,
    ) -> AuthorizedAccess:
        """Verifies if the authorized token is valid and grants access to the specified resource.

        The token is verfied for its validity and - if provided - if it has the specified permission.

        Args:
            token: Token (session or API) to verify.
            resource_name (optional): The access verification will check if the authroized subject is granted access to this resource.
                If `None`, only the token will be checked for validity.
            access_level (optional): The access verification will check if the authroized subject is allowed to access the resource at this level.
                The access level has to be provided if the resource_name is used.

        Raises:
            PermissionDeniedError: If the requested permission is denied.
            UnauthenticatedError: If the token is invalid or expired.

        Returns:
            AuthorizedAccess: Information about the granted permission, authenticated user, and the token.
        """

        # TODO: dynamic access level check: If `None`, the highest access level that the subject is granted for the given resource will be determined.

        permission = None
        if resource_name:
            # Only check for permission if resource name is provided
            # The access level should be provided
            assert access_level is not None
            processed_resource_name = (
                resource_name.rstrip("/").rstrip(":").lstrip("/").lstrip(":")
            )
            permission = (
                processed_resource_name
                + auth_utils.PERMISSION_SEPERATOR
                + access_level.value
            )
        authorized_access = self.get_auth_manager().verify_access(token, permission)
        # Set the authroized access into the request state
        self.request_state.authorized_access = authorized_access
        return authorized_access

    def get_project_manager(self) -> ProjectManager:
        """Returns a Project Manager instance."""
        if not self._project_manager:
            self._project_manager = ProjectManager(self)

        assert self._project_manager is not None
        return self._project_manager

    def get_auth_manager(self) -> AuthManager:
        """Returns an Auth Manager instance."""
        if not self._auth_manager:
            self._auth_manager = AuthManager(self)

        assert self._auth_manager is not None
        return self._auth_manager

    def get_system_manager(self) -> SystemManager:
        """Returns a System Manager instance."""
        if not self._system_manager:
            self._system_manager = SystemManager(self)

        assert self._system_manager is not None
        return self._system_manager

    def get_extension_manager(self) -> ExtensionManager:
        """Returns an Extension Manager instance."""
        if not self._extension_manager:
            self._extension_manager = ExtensionManager(self)
        return self._extension_manager

    def get_json_db_manager(self) -> JsonDocumentOperations:
        """Returns a JSON DB Manager instance."""
        if not self._json_db_manager:
            from contaxy.managers.json_db.postgres import PostgresJsonDocumentManager

            self._json_db_manager = PostgresJsonDocumentManager(
                self.global_state, self.request_state
            )
        return self._json_db_manager

    def get_file_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> FileOperations:
        """Returns a File Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """

        if extension_id is not None and extension_id != CORE_EXTENSION_ID:
            # Request is forwarded to extension
            return self.get_extension_manager().get_extension_client(extension_id)

        if not self._file_manager:
            self._file_manager = self._create_file_manager()
        return self._file_manager

    def _create_file_manager(self) -> FileOperations:
        if self.global_state.settings.S3_ENDPOINT:
            logger.debug("Configuration S3_ENDPOINT set. Using external S3 storage.")
            from contaxy.managers.file.minio import MinioFileManager

            return MinioFileManager(self)
        elif self.global_state.settings.AZURE_BLOB_CONNECTION_STRING:
            logger.debug(
                "Configuration AZURE_BLOB_CONNECTION_STRING set. Using external Azure Blob storage."
            )
            from contaxy.managers.file.azure_blob import AzureBlobFileManager

            return AzureBlobFileManager(self)
        else:
            logger.debug(
                "No external object storage configured. Using internal Minio service."
            )
            raise NotImplementedError(
                "Internal Minio object storage is not implemented! Please configure S3_ENDPOINT or AZURE_BLOB_CONNECTION_STRING."
            )

    def _get_deployment_manager(self) -> DeploymentOperations:
        # Lazyload deployment manager
        if not self._deployment_manager:
            if (
                self.global_state.settings.DEPLOYMENT_MANAGER
                == config.DeploymentManager.DOCKER
            ):
                from contaxy.managers.deployment.docker import DockerDeploymentPlatform

                # Add DB persistence to docker deployment manager
                self._deployment_manager = DeploymentManager(
                    DockerDeploymentPlatform(), self
                )
            elif (
                self.global_state.settings.DEPLOYMENT_MANAGER
                == config.DeploymentManager.KUBERNETES
            ):
                from contaxy.managers.deployment.kubernetes import (
                    KubernetesDeploymentPlatform,
                )

                # Add DB persistence to kubernetes deployment manager
                self._deployment_manager = DeploymentManager(
                    KubernetesDeploymentPlatform(), self
                )

        assert self._deployment_manager is not None
        return self._deployment_manager

    def get_job_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> JobOperations:
        """Returns a Job Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """
        if extension_id is not None and extension_id != CORE_EXTENSION_ID:
            # Request is forwarded to extension
            return self.get_extension_manager().get_extension_client(extension_id)

        return self._get_deployment_manager()

    def get_service_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> ServiceOperations:
        """Returns a Service Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """
        if extension_id is not None and extension_id != CORE_EXTENSION_ID:
            # Request is forwarded to extension
            return self.get_extension_manager().get_extension_client(extension_id)

        return self._get_deployment_manager()

    def get_seed_manager(self) -> SeedOperations:
        if not self._seed_manager:
            self._seed_manager = SeedManager(self)
        return self._seed_manager


def install_components() -> None:
    """Currently only a mock implementation."""

    # create a mocked request for initializing the component manager
    import addict

    mocked_request = addict.Dict()
    # Load and add platform settings to
    global_state = GlobalState(mocked_request.app.state)  # type: ignore
    global_state.settings = config.settings

    # Initialize component manager
    component_manager = ComponentManager(mocked_request)  # type: ignore

    if not global_state.settings.POSTGRES_CONNECTION_URI:
        # Deploy postgres service in system project
        # Set database to use the system namespace
        POSTGRES_DB = global_state.settings.SYSTEM_NAMESPACE
        POSTGRES_PORT = "5432"
        # TODO: Generate credentials?
        POSTGRES_USER = "admin"
        POSTGRES_PASSWORD = "admin"

        service = component_manager.get_service_manager().deploy_service(
            config.SYSTEM_INTERNAL_PROJECT,
            ServiceInput(
                display_name="postgres",
                container_image="postgres",
                endpoints=[POSTGRES_PORT],
                parameters={
                    "POSTGRES_USER": POSTGRES_USER,
                    "POSTGRES_PASSWORD": POSTGRES_PASSWORD,
                    "POSTGRES_DB": POSTGRES_DB,
                },
            ),
        )

        global_state.settings.POSTGRES_CONNECTION_URI = PostgresDsn(
            f"postgres://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{service.id}:{POSTGRES_PORT}/{POSTGRES_DB}"
        )

    # TODO: Do other configurations

    # Close states
    component_manager.close()
    global_state.close()



import ipaddress
import os
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

import docker
import docker.errors
import docker.models.containers
import docker.models.networks
import docker.types
import psutil
from docker import DockerClient
from loguru import logger

from contaxy.config import settings
from contaxy.managers.deployment.utils import (
    _MIN_MEMORY_DEFAULT_MB,
    DEFAULT_DEPLOYMENT_ACTION_ID,
    NO_LOGS_MESSAGE,
    Labels,
    get_gpu_info,
    get_label_string,
    get_network_name,
    get_project_selection_labels,
    get_volume_name,
    map_labels,
    map_list_to_string,
)
from contaxy.schema import ResourceAction
from contaxy.schema.deployment import (
    Deployment,
    DeploymentCompute,
    DeploymentStatus,
    DeploymentType,
    Job,
    JobInput,
    Service,
    ServiceInput,
)
from contaxy.schema.exceptions import (
    ClientBaseError,
    ResourceNotFoundError,
    ServerBaseError,
)

# we create networks in the range of 172.33-255.0.0/24
# Docker by default uses the range 172.17-32.0.0, so we should be save using that range
from contaxy.utils.utils import remove_none_values_from_dict

INITIAL_CIDR_FIRST_OCTET = 10
INITIAL_CIDR_SECOND_OCTET = 0
INITIAL_CIDR = f"{INITIAL_CIDR_FIRST_OCTET}.{INITIAL_CIDR_SECOND_OCTET}.0.0/24"

system_cpu_count = psutil.cpu_count()
system_memory_in_mb = round(psutil.virtual_memory().total / 1024 / 1024, 1)
system_gpu_count = get_gpu_info()


def map_container(
    container: docker.models.containers.Container,
) -> Dict[str, Any]:
    mapped_labels = map_labels(container.labels)

    host_config = container.attrs["HostConfig"]
    compute_resources = DeploymentCompute(
        max_cpus=host_config["NanoCpus"] / 1e9,
        max_memory=host_config["Memory"] / 1000 / 1000,
        max_gpus=None,  # TODO: fill with sensible information - where to get it from?
        min_lifetime=mapped_labels.min_lifetime or 0,
        volume_path=mapped_labels.volume_path,
        # TODO: add max_volume_size, max_replicas
    )

    try:
        container_status = container.status
        if container_status == "created":
            container_status = DeploymentStatus.PENDING.value
        elif container_status == "exited":
            exit_code = container.attrs.get("State", {}).get("ExitCode", 0)
            if exit_code == 0:
                container_status = DeploymentStatus.SUCCEEDED.value
            else:
                container_status = DeploymentStatus.FAILED.value
        status = DeploymentStatus(container_status).value
    except ValueError:
        status = DeploymentStatus.UNKNOWN.value

    started_at = container.attrs.get("State", {}).get("StartedAt", None)
    stopped_at = container.attrs.get("State", {}).get("FinishedAt", None)

    def transform_env_list(envs: list) -> Dict[str, str]:
        transformed_envs = {}
        for env in envs:
            split_env = env.split("=")
            transformed_envs[split_env[0]] = split_env[1] if len(split_env) == 2 else ""
        return transformed_envs

    parameters = transform_env_list(
        container.attrs.get("Config", {}).get("Env", [])
    )  # dict([x for x in [container_env.split("=") for container_env in container_envs]])
    image_tags = container.image.tags
    # Use the first image tag as name. If the image has no tags, use the id.
    if len(image_tags) > 0:
        container_image = image_tags[0]
    else:
        container_image = container.image.short_id

    return {
        "container_image": container_image,
        "command": container.attrs["Config"].get("Entrypoint", []),
        "args": container.attrs["Config"].get("Cmd", []),
        "compute": compute_resources,
        "metadata": mapped_labels.metadata,
        "deployment_type": mapped_labels.deployment_type,
        "description": mapped_labels.description,
        "display_name": mapped_labels.display_name,
        "endpoints": mapped_labels.endpoints,
        #         "exit_code": container.attrs.get("State", {}).get("ExitCode", -1),
        "icon": mapped_labels.icon,
        "id": container.name,
        "internal_id": container.id,
        "parameters": parameters,
        "started_at": started_at,
        "status": status,
        "stopped_at": stopped_at,
    }


def map_service(
    container: docker.models.containers.Container,
) -> Service:
    transformed_container = map_container(container=container)
    transformed_container = remove_none_values_from_dict(transformed_container)

    return Service(**transformed_container)


def map_job(container: docker.models.containers.Container) -> Job:
    transformed_container = map_container(container=container)
    transformed_container = remove_none_values_from_dict(transformed_container)
    return Job(**transformed_container)


# TODO: copied from ML Hub
def create_network(
    client: DockerClient, name: str, labels: Dict[str, str]
) -> docker.models.networks.Network:
    """Create a new network to put the new container into it.

    Containers are separated by networks to prevent them from seeing each other.
    Determine whether a new subnet has to be used. Otherwise, the default Docker subnet would be used
    and, as a result, the amount of networks that can be created is strongly limited.
    We create networks in the range of 172.33-255.0.0/24 whereby Docker by default uses the range 172.17-32.0.0
    See: https://stackoverflow.com/questions/41609998/how-to-increase-maximum-docker-network-on-one-server ; https://loomchild.net/2016/09/04/docker-can-create-only-31-networks-on-a-single-machine/

    Args:
        client (DockerClient): docker client that provides access to the docker API
        name (str): name of the network to be created
        labels (Dict[str, str]): labels that will be attached to the network
    Raises:
        docker.errors.APIError: Thrown by `docker.client.networks.create` upon error.

    Returns:
        docker.Network: the newly created network or the existing network with the given name

    """

    networks = client.networks.list()
    highest_cidr = ipaddress.ip_network(INITIAL_CIDR)

    # determine subnet for the network to be created by finding the highest subnet so far.
    # E.g. when you have three subnets 172.33.1.0, 172.33.2.0, and 172.33.3.0, highest_cidr will be 172.33.3.0
    for network in networks:
        if network.name.lower() == name.lower():
            logger.info(f"Network {name} already exists")
            return network

        has_all_properties = (
            network.attrs["IPAM"]
            and network.attrs["IPAM"]["Config"]
            and len(network.attrs["IPAM"]["Config"]) > 0
            and network.attrs["IPAM"]["Config"][0]["Subnet"]
        )
        if has_all_properties:
            cidr = ipaddress.ip_network(network.attrs["IPAM"]["Config"][0]["Subnet"])

            if (
                cidr.network_address.packed[0] == INITIAL_CIDR_FIRST_OCTET
                and cidr.network_address.packed[1] >= INITIAL_CIDR_SECOND_OCTET
                and cidr > highest_cidr
            ):
                highest_cidr = cidr

    # take the highest cidr and add 256 bits, so that if the highest subnet was 172.33.2.0, the new subnet is 172.33.3.0
    next_cidr = ipaddress.ip_network(
        (highest_cidr.network_address + 256).exploded + "/24"
    )
    if next_cidr.network_address.packed[0] > INITIAL_CIDR_FIRST_OCTET:
        raise RuntimeError("No more possible subnet addresses exist")

    logger.info(f"Create network {name} with subnet {next_cidr.exploded}")
    ipam_pool = docker.types.IPAMPool(
        subnet=next_cidr.exploded, gateway=(next_cidr.network_address + 1).exploded
    )
    ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool])

    return client.networks.create(
        name=name, check_duplicate=True, ipam=ipam_config, labels=labels
    )


def handle_network(
    client: docker.client, project_id: str
) -> docker.models.networks.Network:
    network_name = get_network_name(project_id)
    try:
        network = client.networks.get(network_id=network_name)
    except docker.errors.NotFound:
        network = create_network(
            client=client,
            name=network_name,
            labels={
                Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                Labels.PROJECT_NAME.value: project_id,
            },
        )
    connect_to_network(get_this_container(client), network)
    return network


def connect_to_network(
    container: docker.models.containers.Container,
    network: docker.models.networks.Network,
) -> None:
    # Note: connecting the backend container to the network will cause a connection drop on Macs, so the first service deploy call will be shown as failed, even though it succeeded. On Linux, this problem does not seem to exist
    if container:
        is_backend_connected_to_network = container.attrs["NetworkSettings"][
            "Networks"
        ].get(network.name, False)
        if not is_backend_connected_to_network:
            try:
                network.connect(container)
            except docker.errors.APIError:
                # Remove the network again as it is not connected to any service.
                network.remove()
                raise RuntimeError(
                    f"Could not connect the {container.name} to the network {network.name}"
                )


def reconnect_to_all_networks(
    client: docker.client,
) -> None:
    """Connects the backend container to all networks that belong to the installation.

    A reconnect is required after the container was recreated (e.g. to update the image) and is therefore
    no longer part of all the networks.
    :param client: Docker client
    """
    networks = get_backend_networks(client)
    for network in networks:
        try:
            connect_to_network(get_this_container(client), network)
        except RuntimeError:
            logger.exception(
                "Error reconnecting backend to a network. "
                "This will cause some services to become unavailable."
            )


def get_backend_networks(client: docker.client) -> List[docker.models.networks.Network]:
    try:
        networks = client.networks.list(
            filters={
                "label": get_label_string(
                    Labels.NAMESPACE.value, settings.SYSTEM_NAMESPACE
                )
            }
        )
    except docker.errors.NotFound:
        raise ServerBaseError("Could not list backend networks.")

    return networks


def get_project_container_selection_labels(
    project_id: str, deployment_type: DeploymentType = DeploymentType.SERVICE
) -> List[str]:
    return [
        get_label_string(selection_label[0], selection_label[1])
        for selection_label in get_project_selection_labels(
            project_id=project_id, deployment_type=deployment_type
        )
    ]


def get_project_containers(
    client: DockerClient,
    project_id: str,
    deployment_type: DeploymentType = DeploymentType.SERVICE,
) -> List[docker.models.containers.Container]:
    labels = get_project_container_selection_labels(
        project_id=project_id, deployment_type=deployment_type
    )

    try:
        containers = client.containers.list(all=True, filters={"label": labels})
    except docker.errors.NotFound:
        raise ServerBaseError(
            f"Could not list Docker containers for project '{project_id}'."
        )

    return containers


def get_project_container(
    client: docker.client,
    project_id: str,
    deployment_id: str,
    deployment_type: DeploymentType = DeploymentType.SERVICE,
) -> docker.models.containers.Container:
    labels = get_project_container_selection_labels(
        project_id=project_id, deployment_type=deployment_type
    )
    labels.append(
        get_label_string(Labels.DEPLOYMENT_ID.value, deployment_id),
    )
    try:
        containers = client.containers.list(all=True, filters={"label": labels})
    except docker.errors.NotFound:
        raise ServerBaseError(
            f"Could not list Docker containers for project '{project_id}' and service '{deployment_id}'."
        )

    if len(containers) == 0:
        raise ResourceNotFoundError(
            f"Could not find service '{deployment_id}' of project '{project_id}'."
        )

    return containers[0]


def get_this_container(
    client: DockerClient,
) -> docker.models.containers.Container:
    """This function returns the Docker container in which this code is running or None if it does not run in a container.

    Args:
        client (docker.client): The Docker client object

    Returns:
        docker.models.containers.Container: If this code runs in a container, it returns this container otherwise None
    """

    # "Detect" whether it runs in a container by checking the following environment variables
    hostname = os.getenv("HOSTNAME", False)
    if not os.getenv("IS_CONTAXY_CONTAINER", False) or hostname is None:
        return None
    return client.containers.get(hostname)


def delete_container(
    client: DockerClient,
    container: docker.models.containers.Container,
    delete_volumes: bool = False,
) -> None:
    try:
        container.stop()
    except docker.errors.APIError:
        pass

    try:
        container.remove(v=delete_volumes)
    except docker.errors.APIError:
        raise ClientBaseError(
            status_code=500,
            message=f"Could not delete deployment '{container.name}'.",
        )

    # Named volumes must be deleted manually
    if delete_volumes:
        for mount in container.attrs.get("Mounts", []):
            if mount.get("Type") == "volume" and "Name" in mount:
                try:
                    client.api.remove_volume(mount["Name"])
                except docker.errors.APIError:
                    raise ClientBaseError(
                        status_code=500,
                        message=f"Could not delete volume {mount['Name']} of deployment '{container.name}'.",
                    )


def check_minimal_resources(
    min_cpus: int,
    min_memory: int,
    min_gpus: int,
    compute_resources: DeploymentCompute = None,
) -> None:
    if min_cpus > system_cpu_count:
        raise RuntimeError(
            f"The minimal amount of cpus of {min_cpus} cannot be fulfilled as the system has only {system_cpu_count} cpus."
        )
    if min_memory > system_memory_in_mb:
        raise RuntimeError(
            f"The minimal amount of memory of {min_memory}MB cannot be fulfilled as the system has only {system_memory_in_mb}MB memory"
        )
    if min_gpus > system_gpu_count:
        raise RuntimeError(
            f"The minimal amount of gpus of {min_gpus} cannot be fulfilled as the system has only {system_gpu_count} gpus."
        )

    if compute_resources is None:
        return

    if compute_resources.max_replicas is not None:
        raise RuntimeError("Replicas are not supported in Docker-mode")


def extract_minimal_resources(
    compute_resources: DeploymentCompute,
) -> Tuple[int, int, int]:
    min_cpus = (
        compute_resources.min_cpus if compute_resources.min_cpus is not None else 0
    )
    min_memory = (
        compute_resources.min_memory if compute_resources.min_memory is not None else 0
    )
    min_gpus = (
        compute_resources.min_gpus if compute_resources.min_gpus is not None else 0
    )

    return min_cpus, min_memory, min_gpus


def define_mounts(
    project_id: str,
    container_name: str,
    compute_resources: DeploymentCompute,
    service_requirements: Optional[List[str]] = [],
) -> list:
    mounts = []
    # if service_requirements and "docker" in service_requirements:
    #     # TODO: IMPORTANT mark container as having extended privileges so that on a higher level the platform
    #     # can prevent that a non-admin creates a container with the Docker socket mounted
    #     mounts.append(
    #         docker.types.Mount(
    #             target="/var/run/docker.sock", source="/var/run/docker.sock"
    #         )
    #     )

    if (
        compute_resources.volume_path is not None
        and compute_resources.volume_path != ""
    ):
        if settings.HOST_DATA_ROOT_PATH is None:
            mount_type = "volume"
            prefix = ""
        else:
            mount_type = "bind"
            prefix = settings.HOST_DATA_ROOT_PATH
        mounts.append(
            docker.types.Mount(
                target=str(compute_resources.volume_path),
                source=f"{prefix}{get_volume_name(project_id, container_name)}",
                labels={
                    Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                    Labels.PROJECT_NAME.value: project_id,
                    Labels.DEPLOYMENT_ID.value: container_name,
                },
                type=mount_type,
            )
        )

    return mounts


def create_container_config(
    deployment: Deployment,
    project_id: str,
) -> Dict[str, Any]:
    compute_resources = deployment.compute
    (
        min_cpus,
        min_memory,
        min_gpus,
    ) = extract_minimal_resources(compute_resources=compute_resources)
    check_minimal_resources(
        min_cpus=min_cpus,
        min_memory=min_memory,
        min_gpus=min_gpus,
    )
    max_cpus = (
        compute_resources.max_cpus if compute_resources.max_cpus is not None else 1
    )
    max_memory = (
        compute_resources.max_memory if compute_resources.max_memory is not None else 6
    )
    # Make sure that the user-entered compute requirements are not bigger than the system's maximum available
    nano_cpus = min(max_cpus, system_cpu_count) * 1e9
    # With regards to memory, Docker requires at least 6MB for a container
    mem_limit = f"{max(_MIN_MEMORY_DEFAULT_MB, min(max_memory, system_memory_in_mb))}MB"

    container_name = deployment.id
    mounts = define_mounts(
        project_id=project_id,
        container_name=container_name,
        compute_resources=compute_resources,
        service_requirements=deployment.requirements,
    )

    return {
        "image": deployment.container_image,
        "entrypoint": deployment.command,
        "command": deployment.args,
        "detach": True,
        "environment": deployment.parameters,
        "labels": {
            **deployment.metadata,
            Labels.DISPLAY_NAME.value: deployment.display_name,
            Labels.DEPLOYMENT_TYPE.value: deployment.deployment_type.value,
            Labels.DESCRIPTION.value: deployment.description,
            Labels.ENDPOINTS.value: map_list_to_string(deployment.endpoints),
            Labels.REQUIREMENTS.value: map_list_to_string(deployment.requirements),
            Labels.ICON.value: deployment.icon,
            Labels.MIN_LIFETIME.value: str(compute_resources.min_lifetime),
            Labels.CREATED_BY.value: deployment.created_by,
            Labels.VOLUME_PATH.value: compute_resources.volume_path,
        },
        "name": container_name,
        "nano_cpus": int(nano_cpus),
        "mem_limit": mem_limit,
        "network": get_network_name(project_id),
        "restart_policy": {"Name": "on-failure", "MaximumRetryCount": 10},
        "mounts": mounts if mounts != [] else None,
    }


def read_container_logs(
    container: docker.models.containers.Container,
    lines: Optional[int] = None,
    since: Optional[datetime] = None,
) -> str:
    try:
        logs = container.logs(tail=lines or "all", since=since)
    except docker.errors.APIError:
        logger.error(f"Could not read logs of container {container.name}.")

    if logs:
        logs = logs.decode("utf-8")
    else:
        logs = NO_LOGS_MESSAGE

    return logs


def wait_for_container(
    container: docker.models.containers.Container,
    client: docker.client,
    timeout: int = 60,
) -> docker.models.containers.Container:
    start = time.time()
    while time.time() - start < timeout:
        container_info = client.containers.get(container.id)
        if container_info.status.lower() != "created":
            return container_info
        else:
            time.sleep(2)

    raise RuntimeError(f"Timeout while waiting for container {container.id}.")


def list_deploy_service_actions(
    project_id: str, deploy_input: Union[ServiceInput, JobInput]
) -> List[ResourceAction]:
    compute_resources = deploy_input.compute or DeploymentCompute()
    min_cpus, min_memory, min_gpus = extract_minimal_resources(compute_resources)
    try:
        check_minimal_resources(
            min_cpus=min_cpus, min_memory=min_memory, min_gpus=min_gpus
        )
    except RuntimeError:
        return []

    return [
        ResourceAction(
            action_id=DEFAULT_DEPLOYMENT_ACTION_ID,
            display_name=DEFAULT_DEPLOYMENT_ACTION_ID,
        )
    ]



import json
from datetime import datetime, timezone
from typing import Dict, List, Optional

import json_merge_patch

from contaxy.operations import JsonDocumentOperations
from contaxy.schema.exceptions import ResourceAlreadyExistsError, ResourceNotFoundError
from contaxy.schema.json_db import JsonDocument
from contaxy.utils.state_utils import GlobalState, RequestState


class InMemoryDictJsonDocumentManager(JsonDocumentOperations):
    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
    ):
        """Initializes the Postgres Json Document Manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
        """
        self._global_state = global_state
        self._request_state = request_state
        self._dict_db: Optional[Dict] = None

    def _get_collection(self, project_id: str, collection_id: str) -> Dict:
        """Lazyloads the specified collection.

        Args:
            project_id: The ID of the project
            collection_id: The ID of the collection.

        Returns:
            Dict: The collection Dict.
        """
        if self._dict_db is None:
            state_namespace = self._global_state[InMemoryDictJsonDocumentManager]
            if not state_namespace.dict_db:
                state_namespace.dict_db = {}
            self._dict_db = state_namespace.dict_db

        assert self._dict_db is not None
        if project_id not in self._dict_db:
            self._dict_db[project_id] = {}

        project_db = self._dict_db[project_id]
        if collection_id not in project_db:
            project_db[collection_id] = {}

        return project_db[collection_id]

    def create_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
        upsert: bool = True,
    ) -> JsonDocument:
        """Creates a JSON document for a given key.

        If a document already exists for the given key, the document will be overwritten.

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) to use to store this JSON document.
            key: Key of the JSON document.
            json_document: The actual JSON document value.
            upsert: If `True`, the document will be updated/overwritten if it already exists.

        Returns:
            JsonDocument: The created JSON document.
        """
        collection = self._get_collection(project_id, collection_id)

        if not upsert and key in collection:
            raise ResourceAlreadyExistsError(
                "A document with the key {key} already exists."
            )

        created_document = JsonDocument(
            key=key,
            json_value=json_document,
            created_at=datetime.now(timezone.utc),
            update_at=datetime.now(timezone.utc),
        )
        collection[key] = created_document.dict()
        return created_document

    def update_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
    ) -> JsonDocument:
        """Updates a JSON document.

        The update is applied on the existing document based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.
            json_document: The actual JSON document value.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.

        Returns:
            JsonDocument: The updated JSON document.
        """
        collection = self._get_collection(project_id, collection_id)

        current_document = self.get_json_document(project_id, collection_id, key)

        updated_json = json_merge_patch.merge(
            json.loads(current_document.json_value), json.loads(json_document)
        )

        current_document.json_value = json.dumps(updated_json)
        current_document.updated_at = datetime.now(timezone.utc)
        collection[key] = current_document.dict()

        return self.get_json_document(project_id, collection_id, key)

    def list_json_documents(
        self,
        project_id: str,
        collection_id: str,
        filter: Optional[str] = None,
        keys: Optional[List[str]] = None,
    ) -> List[JsonDocument]:
        """Lists all JSON documents for the given project collection.

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            filter (optional): Allows to filter the result documents based on a JSONPath expression ([JSON Path Specification](https://goessner.net/articles/JsonPath/)). The filter is only applied to filter documents in the list. It is not usable to extract specific properties.
            keys (optional): Json Document Ids, i.e. DB row keys. Defaults to `None`.

        Returns:
            List[JsonDocument]: List of JSON documents.
        """
        # TODO: support keys parameter
        collection = self._get_collection(project_id, collection_id)
        documents: List[JsonDocument] = []
        for doc_key in collection.keys():
            documents.append(JsonDocument(**collection[doc_key]))

        if filter:
            # TODO: filter currently not working since json path of postgres is different than the impl below
            # TODO: just return all data -> implmenetation needs to take on filtering
            pass
            # Filter based on jsonpath
            # filtered_documents: List[JsonDocument] = []
            # jsonpath_expr = jsonpath_ng.ext.parse(filter)
            # for document in documents:
            #    if [
            #        match.value
            #        for match in jsonpath_expr.find([json.loads(document.json_value)])
            #    ]:
            #        filtered_documents.append(document)
            # documents = filtered_documents

        return documents

    def get_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
    ) -> JsonDocument:
        """Returns a single JSON document.

        Args:
            project_id: Project ID associated with the JSON document.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.

        Returns:
            JsonDocument: A JSON document.
        """
        collection = self._get_collection(project_id, collection_id)
        if key not in collection:
            raise ResourceNotFoundError(
                f"The json document with the key {key} does not exist."
            )

        return JsonDocument(**collection[key])

    def delete_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
    ) -> None:
        """Deletes a single JSON document.

        If no other document exists in the project collection, the collection will be deleted.

        Args:
            project_id: Project ID associated with the JSON document.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.
        """
        collection = self._get_collection(project_id, collection_id)
        if key not in collection:
            raise ResourceNotFoundError(
                f"The json document with the key {key} does not exists."
            )
        del collection[key]

    def delete_json_collection(
        self,
        project_id: str,
        collection_id: str,
    ) -> None:
        # Lazy load if not already happend
        self._get_collection(project_id, collection_id)
        # Empty out collection
        if self._dict_db and self._dict_db[project_id]:
            self._dict_db[project_id][collection_id] = {}

    def delete_json_collections(self, project_id: str) -> None:
        state_namespace = self._global_state[InMemoryDictJsonDocumentManager]
        if state_namespace.dict_db:
            state_namespace.dict_db[project_id] = {}

        if self._dict_db:
            self._dict_db[project_id] = {}  # type: ignore



import os
from typing import Any, List, Optional

from fastapi import APIRouter, Body, Depends, Form, Query, Request, status
from requests_oauthlib import OAuth2Session
from starlette.responses import RedirectResponse, Response

from contaxy import config
from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.managers.auth import AuthManager
from contaxy.schema import (
    ApiToken,
    CoreOperations,
    ExtensibleOperations,
    OAuth2TokenRequestFormNew,
    OAuthToken,
    OAuthTokenIntrospection,
    TokenType,
)
from contaxy.schema.auth import AccessLevel, OAuth2ErrorDetails, OAuth2TokenGrantTypes
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
    ClientBaseError,
    ServerBaseError,
)
from contaxy.schema.system import SystemState
from contaxy.utils import auth_utils, id_utils

OAUTH_CALLBACK_ROUTE = "auth/oauth/callback"

router = APIRouter(tags=["auth"], responses={**VALIDATION_ERROR_RESPONSE})


@router.get(
    "/auth/login-page",
    operation_id=ExtensibleOperations.OPEN_LOGIN_PAGE.value,
    summary="Open the login page.",
    status_code=status.HTTP_303_SEE_OTHER,
)
def open_login_page(
    request: Request,
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Returns or redirect to the login-page."""
    if not config.settings.OIDC_AUTH_URL or not config.settings.OIDC_CLIENT_ID:
        raise ServerBaseError(
            "External OIDC Provider is not configured. Please set OIDC_AUTH_URL and further relevant OIDC config params."
        )

    # Needed for test setups
    schema = "http://" if os.getenv("OAUTHLIB_INSECURE_TRANSPORT") else "https://"
    callback_uri = os.path.join(
        schema,
        config.settings.get_redirect_uri(),
        config.settings.CONTAXY_API_PATH,
        OAUTH_CALLBACK_ROUTE,
    )
    session = OAuth2Session(
        config.settings.OIDC_CLIENT_ID,
        scope=["openid", "email"],
        redirect_uri=callback_uri,
    )
    url, state = session.authorization_url(config.settings.OIDC_AUTH_URL)
    # TODO: Check if connector id is required here
    redirect_response = RedirectResponse(url, status_code=status.HTTP_303_SEE_OTHER)
    redirect_response.set_cookie(key="oauth_state", value=state, httponly=True)
    return redirect_response


@router.get(
    "/auth/login",
    operation_id=CoreOperations.LOGIN_USER_SESSION.value,
    summary="Login a user session.",
    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
)
def login_user_session(
    username: str = Form(
        ..., description="The user’s username or email used for login."
    ),
    password: str = Form(..., description="The user’s password."),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Signs in the user based on username and password credentials.

    This will set http-only cookies containg tokens with full user access.
    """
    system_info = component_manager.get_system_manager().get_system_info()
    if system_info.system_state != SystemState.RUNNING:
        raise ClientBaseError(
            status.HTTP_400_BAD_REQUEST,
            "Login is not possible before system has been initialized!",
        )

    oauth_token = component_manager.get_auth_manager().request_token(
        OAuth2TokenRequestFormNew(
            grant_type=OAuth2TokenGrantTypes.PASSWORD,
            username=username,
            password=password,
            scope=auth_utils.construct_permission(
                "*", AccessLevel.ADMIN
            ),  # Get full scope
        )
    )

    # TODO: set the right start page of webapp
    response = RedirectResponse(
        url="/app", status_code=status.HTTP_307_TEMPORARY_REDIRECT
    )
    # TODO: Set cookie and path or other configurations
    token_info = component_manager.get_auth_manager().introspect_token(
        oauth_token.access_token
    )
    response.set_cookie(
        key=config.API_TOKEN_NAME, value=oauth_token.access_token, httponly=True
    )
    if token_info.sub:
        # Set user ID as cookie as well
        response.set_cookie(
            key=config.AUTHORIZED_USER_COOKIE,
            value=id_utils.extract_user_id_from_resource_name(token_info.sub),
        )

    return response


@router.get(
    "/auth/logout",
    operation_id=CoreOperations.LOGOUT_USER_SESSION.value,
    summary="Logout a user session.",
    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
)
def logout_user_session(
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Removes all session token cookies and redirects to the login page.

    When making requests to the this endpoint, the browser should be redirected to this endpoint.
    """
    # RedirectResponse("./login", status_code=307)
    return component_manager.get_auth_manager().logout_session()


@router.get(
    "/auth/tokens",
    operation_id=CoreOperations.LIST_API_TOKENS.value,
    response_model=List[ApiToken],
    response_model_exclude_unset=True,
    summary="List API tokens.",
    status_code=status.HTTP_200_OK,
    responses={**AUTH_ERROR_RESPONSES},
)
def list_api_tokens(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns list of created API tokens associated with the authenticated user."""
    authorized_access = component_manager.verify_access(token)
    # Check if the caller has admin access on the user resource
    component_manager.verify_access(
        token,
        authorized_access.authorized_subject,
        AccessLevel.ADMIN,
    )

    return component_manager.get_auth_manager().list_api_tokens(
        authorized_access.authorized_subject
    )


@router.get(
    "/auth/permissions",
    summary="Lists permissions for given resource.",
    operation_id=CoreOperations.LIST_RESOURCE_PERMISSIONS.value,
    status_code=status.HTTP_200_OK,
    response_model=List[str],
)
def list_resource_permissions(
    resource_name: str,
    resolve_roles: bool = True,
    use_cache: bool = True,
    token: str = Depends(get_api_token),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Lists the permissions for a given resource, admin access for that resource is required."""
    component_manager.verify_access(token, resource_name, AccessLevel.ADMIN)

    return component_manager.get_auth_manager().list_permissions(
        resource_name, resolve_roles, use_cache
    )


@router.post(
    "/auth/permissions",
    summary="Add permission to specified resource.",
    operation_id=CoreOperations.SET_RESOURCE_PERMISSIONS.value,
    status_code=status.HTTP_204_NO_CONTENT,
)
def add_resource_permission(
    resource_name: str,
    permission: str,
    token: str = Depends(get_api_token),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Adds permission to specified resource, admin access to the /auth/permissions resource is required."""
    component_manager.verify_access(token, "/auth/permissions", AccessLevel.ADMIN)

    component_manager.get_auth_manager().add_permission(resource_name, permission)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/auth/permissions",
    summary="Deletes permission of specified resource.",
    operation_id=CoreOperations.DELETE_RESOURCE_PERMISSIONS.value,
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_resource_permissions(
    resource_name: str,
    permission: str,
    remove_sub_permissions: bool = False,
    token: str = Depends(get_api_token),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Deletes permission of specified resource, admin access to the /auth/permissions resource is required."""
    component_manager.verify_access(token, "/auth/permissions", AccessLevel.ADMIN)

    component_manager.get_auth_manager().remove_permission(
        resource_name, permission, remove_sub_permissions
    )
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/auth/resources",
    summary="List all resources that have a certain permission.",
    operation_id=CoreOperations.GET_RESOURCES_WITH_PERMISSION.value,
    status_code=status.HTTP_200_OK,
    response_model=List[str],
)
def list_resources_with_permission(
    permission: str,
    resource_name_prefix: str,
    token: str = Depends(get_api_token),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """List all resources that have a certain permission, admin access to the /auth/resources resource is required."""
    component_manager.verify_access(token, "/auth/resources", AccessLevel.ADMIN)

    component_manager.get_auth_manager().list_resources_with_permission(
        permission, resource_name_prefix
    )


@router.post(
    "/auth/tokens",
    operation_id=CoreOperations.CREATE_TOKEN.value,
    response_model=str,
    summary="Create API or session token.",
    status_code=status.HTTP_200_OK,
    responses={**AUTH_ERROR_RESPONSES},
)
def create_token(
    scopes: Optional[List[str]] = Query(
        None,
        title="Scopes",
        alias="scope",
        description="Scopes requested for this token. If none specified, the token will be generated with same set of scopes as the authorized token.",
    ),
    token_type: TokenType = Query(
        TokenType.SESSION_TOKEN,
        description="Type of the token.",
        type="string",
    ),
    description: Optional[str] = Query(
        None, description="Attach a short description to the generated token."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns a session or API token with the specified scopes.

    If no scopes are specified, the token will be generated with the same scopes as the authorized token.

    The API token can be deleted (revoked) at any time.
    In comparison, the session token cannot be revoked but expires after a short time (a few minutes).

    This operation can only be called with API tokens (or refresh tokens) due to security aspects.
    Session tokens are not allowed to create other tokens.
    Furthermore, tokens can only be created if the API token used for authorization is granted at least
    the same access level on the given resource. For example, a token with `write` access level on a given resource
    allows to create new tokens with `write` or `read` granted level on that resource.
    """
    authorized_access = component_manager.verify_access(token)

    if not scopes:
        # Get scopes from token
        scopes = []
        token_introspection = component_manager.get_auth_manager().introspect_token(
            token
        )
        if token_introspection.scope:
            scopes = token_introspection.scope.split()

    # Check that caller has requested scopes/permissions
    for scope in scopes:
        resource_name, access_level = auth_utils.parse_permission(scope)
        component_manager.verify_access(token, resource_name, access_level)

    return component_manager.get_auth_manager().create_token(
        token_subject=authorized_access.authorized_subject,
        scopes=scopes,
        token_type=token_type,
        description=description,
    )


@router.post(
    "/auth/tokens/verify",
    operation_id=CoreOperations.VERIFY_ACCESS.value,
    summary="Verify a Session or API Token.",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={**AUTH_ERROR_RESPONSES},
)
def verify_access(
    token_in_body: Optional[str] = Body(
        None,
        title="Token",
        description="Token to verify.",
    ),
    permission: Optional[str] = Query(
        None,
        title="Resource Type ",
        description="The token is checked if it is granted this permission. If none specified, only the existence or validity of the token itself is checked.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Verifies a session or API token for its validity and - if provided - if it has the specified permission.

    Returns an successful HTTP Status code if verification was successful, otherwise an error is returned.
    """
    if token_in_body:
        # Prefer token from request body
        token = token_in_body
    component_manager.get_auth_manager().verify_access(token, permission)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


# OAuth Endpoints
def _add_cookies_to_response(
    auth_manager: AuthManager, response: Response, token: OAuthToken
) -> Response:

    # TODO: Set cookie and path or other configurations
    token_info = auth_manager.introspect_token(token.access_token)
    response.set_cookie(
        key=config.API_TOKEN_NAME, value=token.access_token, httponly=True
    )
    if token_info.sub:
        # Set user ID as cookie as well
        response.set_cookie(
            key=config.AUTHORIZED_USER_COOKIE,
            value=id_utils.extract_user_id_from_resource_name(token_info.sub),
        )

    return response


@router.post(
    "/auth/oauth/token",
    operation_id=CoreOperations.REQUEST_TOKEN.value,
    response_model=OAuthToken,
    summary="Request a token (OAuth2 Endpoint).",
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
    responses={
        status.HTTP_400_BAD_REQUEST: {"model": OAuth2ErrorDetails},
    },
)
def request_token(
    token_request_form: OAuth2TokenRequestFormNew = Depends(
        OAuth2TokenRequestFormNew.as_form  # type: ignore
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Returns an access tokens, ID tokens, or refresh tokens depending on the request parameters.

     The token endpoint is used by the client to obtain an access token by
     presenting its authorization grant or refresh token.

     The token endpoint supports the following grant types:
     - [Password Grant](https://tools.ietf.org/html/rfc6749#section-4.3.2): Used when the application exchanges the user’s username and password for an access token.
         - `grant_type` must be set to `password`
         - `username` (required): The user’s username.
         - `password` (required): The user’s password.
         - `scope` (optional): Optional requested scope values for the access token.
     - [Refresh Token Grant](https://tools.ietf.org/html/rfc6749#section-6): Allows to use refresh tokens to obtain new access tokens.
         - `grant_type` must be set to `refresh_token`
         - `refresh_token` (required): The refresh token previously issued to the client.
         - `scope` (optional): Requested scope values for the new access token. Must not include any scope values not originally granted by the resource owner, and if omitted is treated as equal to the originally granted scope.
     - [Client Credentials Grant](https://tools.ietf.org/html/rfc6749#section-4.4.2): Request an access token using only its client
    credentials.
         - `grant_type` must be set to `client_credentials`
         - `scope` (optional): Optional requested scope values for the access token.
         - Client Authentication required (e.g. via client_id and client_secret or auth header)
     - [Authorization Code Grant](https://tools.ietf.org/html/rfc6749#section-4.1): Used to obtain both access tokens and refresh tokens based on an authorization code from the `/authorize` endpoint.
         - `grant_type` must be set to `authorization_code`
         - `code` (required): The authorization code that the client previously received from the authorization server.
         - `redirect_uri` (required): The redirect_uri parameter included in the original authorization request.
         - Client Authentication required (e.g. via client_id and client_secret or auth header)

    For password, client credentials, and refresh token flows, calling this endpoint is the only step of the flow.
    For the authorization code flow, calling this endpoint is the second step of the flow.

    This endpoint implements the [OAuth2 Token Endpoint](https://tools.ietf.org/html/rfc6749#section-3.2).
    """
    system_info = component_manager.get_system_manager().get_system_info()
    if system_info.system_state != SystemState.RUNNING:
        raise ClientBaseError(
            status.HTTP_400_BAD_REQUEST,
            "Login is not possible before system has been initialized!",
        )

    oauth_token = component_manager.get_auth_manager().request_token(token_request_form)
    if token_request_form.set_as_cookie:
        response = Response(status_code=status.HTTP_200_OK)
        return _add_cookies_to_response(
            component_manager.get_auth_manager(), response, oauth_token
        )
    return oauth_token


@router.post(
    "/auth/oauth/revoke",
    operation_id=CoreOperations.REVOKE_TOKEN.value,
    summary="Revoke a token (OAuth2 Endpoint).",
    status_code=status.HTTP_200_OK,
    responses={
        status.HTTP_400_BAD_REQUEST: {"model": OAuth2ErrorDetails},
    },
)
def revoke_token(
    token: str = Form(..., description="The token that should be revoked."),
    token_type_hint: Optional[str] = Form(
        None,
        description="A hint about the type of the token submitted for revokation.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Revokes a given token.

    This will delete the API token, preventing further requests with the given token.
    Because of caching, the API token might still be usable under certain conditions
    for some operations for a maximum of 15 minutes after deletion.

    This endpoint implements the OAuth2 Revocation Flow ([RFC7009](https://tools.ietf.org/html/rfc7009)).
    """
    # TODO: Implement authentication required by spec
    component_manager.get_auth_manager().revoke_token(token)


@router.post(
    "/auth/oauth/introspect",
    operation_id=CoreOperations.INTROSPECT_TOKEN.value,
    response_model=OAuthTokenIntrospection,
    summary="Introspect a token (OAuth2 Endpoint).",
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
    responses={**AUTH_ERROR_RESPONSES},
)
def introspect_token(
    token: str = Form(..., description="The token that should be instrospected."),
    token_type_hint: Optional[str] = Form(
        None,
        description="A hint about the type of the token submitted for introspection (e.g. `access_token`, `id_token` and `refresh_token`).",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Introspects a given token.

    Returns a boolean that indicates whether it is active or not.
    If the token is active, additional data about the token is also returned.
    If the token is invalid, expired, or revoked, it is considered inactive.

    This endpoint implements the [OAuth2 Introspection Flow](https://www.oauth.com/oauth2-servers/token-introspection-endpoint/) ([RFC7662](https://tools.ietf.org/html/rfc7662)).
    """
    # TODO: Implement authentication required by spec
    return component_manager.get_auth_manager().introspect_token(token)


@router.get(
    "/auth/oauth/callback",
    operation_id=CoreOperations.LOGIN_CALLBACK.value,
    summary="Open the login page (OAuth2 Client Endpoint).",
    status_code=status.HTTP_200_OK,
)
def login_callback(
    code: str = Query(
        ..., description="The authorization code generated by the authorization server."
    ),
    state: Optional[str] = Query(None),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Callback to finish the login process (OAuth2 Client Endpoint).

    The authorization `code` is exchanged for an access and ID token.
    The ID token contains all relevant user information and is used to login the user.
    If the user does not exist, a new user will be created with the information from the ID token.

    Finally, the user is redirected to the webapp and a session/refresh token is set in the cookies.

    This endpoint implements the [Authorization Response](https://tools.ietf.org/html/rfc6749#section-4.1.2) from RFC6749.
    """
    system_info = component_manager.get_system_manager().get_system_info()
    if system_info.system_state != SystemState.RUNNING:
        raise ClientBaseError(
            status.HTTP_400_BAD_REQUEST,
            "Login is not possible before system has been initialized!",
        )

    auth_manager = component_manager.get_auth_manager()
    # Needed for test setups
    schema = "http://" if os.getenv("OAUTHLIB_INSECURE_TRANSPORT") else "https://"
    oauth_token = auth_manager.login_callback(
        code,
        os.path.join(
            schema,
            config.settings.get_redirect_uri(),
            config.settings.CONTAXY_API_PATH,
            OAUTH_CALLBACK_ROUTE,
        ),
        component_manager.get_project_manager(),
        state,
    )

    # ! Currently, the webapp needs to be accessible via the same host/port
    response = RedirectResponse(
        os.path.join(
            schema,
            config.settings.get_redirect_uri(),
            f"{config.settings.CONTAXY_WEBAPP_PATH}/",  # TODO remove when nginx routing is possible without trailing slash
        ),
        status_code=status.HTTP_303_SEE_OTHER,
    )

    return _add_cookies_to_response(auth_manager, response, oauth_token)


@router.get(
    "/auth/oauth/enabled",
    operation_id=CoreOperations.OAUTH_ENABLED.value,
    summary="Check if external OAuth2 (OIDC) IDP is enabled.",
    status_code=status.HTTP_200_OK,
    response_model=str,
)
def is_external_idp_enabled() -> Any:
    """Returns the value of `OIDC_AUTH_ENABLED`.

    Returns "0" if it is not set, "1" if it is set.
    """
    return Response(
        status_code=status.HTTP_200_OK,
        content=str(int(config.settings.OIDC_AUTH_ENABLED)),
    )



from typing import Any, Dict, Mapping, Optional, Union

from fastapi import HTTPException, status
from pydantic import BaseModel, Field


# Inspired by: https://cloud.google.com/apis/design/errors
class ProblemDetails(BaseModel):
    code: Optional[int] = Field(
        None,
        ge=100,
        lt=600,
        example=400,
        description="The HTTP status code generated by the origin server for this occurrence of the problem.",
    )
    message: Optional[str] = Field(
        None,
        # example="Service Unavailable",
        description="A developer-facing human-readable error message in English.",
    )
    explanation: Optional[str] = Field(
        None,
        # example="Service Unavailable",
        description="A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.",
    )
    details: Optional[Any] = Field(
        None,
        description="A map of additional problem details.",
    )


# Based on RFC 7807 Standard: https://tools.ietf.org/html/rfc7807
# https://opensource.zalando.com/restful-api-guidelines/problem-1.0.1.yaml
# class ProblemDetailsFormat(BaseModel):
#     typ(e): str = Field(
#         "about:blank",
#         # example="problem/connection-error",
#         description="A relative URI reference that uniquely identifies the problem type only in the context of the provided API. In combination with the API endpoint, this might point to further documentation.",
#     )
#     status: Optional[int] = Field(
#         None,
#         ge=100,
#         lt=600,
#         # example=503,
#         description="The HTTP status code generated by the origin server for this occurrence of the problem.",
#     )
#     title: Optional[str] = Field(
#         None,
#         # example="Service Unavailable",
#         description="A short summary of the problem type. Written in English and readable for engineers, usually not suited for non technical stakeholders and not localized.",
#     )
#     detail: Optional[str] = Field(
#         None,
#         # example="Connection to database timed out",
#         description="A human readable explanation specific to this occurrence of the problem that is helpful to locate the problem and give advice on how to proceed. Written in English and readable for engineers, usually not suited for non technical stakeholders and not localized.",
#     )
#     instance: Optional[str] = Field(
#         None,
#         description="A relative URI reference that identifies the specific occurrence of the problem. It may or may not yield further information if dereferenced.",
#     )
#     metadata: Optional[Dict] = Field(
#         None,
#         description="A map of additional problem details/metadata.",
#     )


class ClientBaseError(HTTPException):
    """Basic exception class for all errors that should be shown to the client/user.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    def __init__(
        self,
        status_code: int,
        message: str,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the exception.

        Args:
            status_code: The HTTP status code associated with the error.
            message: A short summary of the error.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """

        super(ClientBaseError, self).__init__(
            status_code=status_code,
            detail=message,
        )

        self.explanation = explanation
        self.metadata = metadata
        self.resource = resource


class ServerBaseError(Exception):
    """Basic exception class for all server internal errors that should not be shown with details to the user.

    If the error is not handled, an `Internal Server Error` (Status Code 500) will be shown
    to the client (user) without any additional details. In this case, the exception will be
    automatically logged.
    """

    def __init__(self, *args: object) -> None:
        """Initializes the exception.

        Args:
            args: A collection of
        """
        super(ServerBaseError, self).__init__(*args)


class UnauthenticatedError(ClientBaseError):
    """Client error that indicates invalid, expired, or missing authentication credentials.

    The error message should contain specific details about the problem, e.g.:

    - No access token was provided.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_401_UNAUTHORIZED
    _DEFAULT_MESSAGE = "Invalid authentication credentials."
    _DEFAULT_EXPLANATION = "This can happen because credentials or token are missing, expired, or otherwise invalid."

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
    ) -> None:
        """Initializes the error.

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
        """
        super(UnauthenticatedError, self).__init__(
            status_code=UnauthenticatedError._HTTP_STATUS_CODE,
            message=message or UnauthenticatedError._DEFAULT_MESSAGE,
            explanation=explanation or UnauthenticatedError._DEFAULT_EXPLANATION,
            metadata=metadata,
        )


class PermissionDeniedError(ClientBaseError):
    """Client error that indicates that a client does not have sufficient permission for the request.

    The error message should contain specific details about the resource, e.g.:

    - Permission 'xxx' denied on resource 'yyy'.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_403_FORBIDDEN
    _DEFAULT_MESSAGE = "The authorized user does not have sufficient permission."
    _DEFAULT_EXPLANATION = "This can happen because the token does not have the right scopes or the user doesn't have the required permission."

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the error.

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """
        super(PermissionDeniedError, self).__init__(
            status_code=PermissionDeniedError._HTTP_STATUS_CODE,
            message=message or PermissionDeniedError._DEFAULT_MESSAGE,
            explanation=explanation or PermissionDeniedError._DEFAULT_EXPLANATION,
            metadata=metadata,
            resource=resource,
        )


class ResourceNotFoundError(ClientBaseError):
    """Client error that indicates that a specified resource is not found.

    The error message should contain specific details about the resource, e.g.:

    - Resource 'xxx' not found.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_404_NOT_FOUND
    _DEFAULT_MESSAGE = "A specified resource is not found."

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the error.

        This error should be raised if

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """
        super(ResourceNotFoundError, self).__init__(
            status_code=ResourceNotFoundError._HTTP_STATUS_CODE,
            message=message or ResourceNotFoundError._DEFAULT_MESSAGE,
            explanation=explanation,
            metadata=metadata,
            resource=resource,
        )


class ResourceAlreadyExistsError(ClientBaseError):
    """Client error that indicates that a resource that a client tried to create already exists.

    The error message should contain specific details about the problem and resource, e.g.:

    - Resource 'xxx' already exists.
    - Couldn’t acquire lock on resource ‘xxx’.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_409_CONFLICT
    _DEFAULT_MESSAGE = "The resource already exists."

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the error.

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """
        super(ResourceAlreadyExistsError, self).__init__(
            status_code=ResourceAlreadyExistsError._HTTP_STATUS_CODE,
            message=message or ResourceAlreadyExistsError._DEFAULT_MESSAGE,
            explanation=explanation,
            metadata=metadata,
        )


class ClientValueError(ClientBaseError, ValueError):
    """Client error that indicates that the client input is invalid.

    The error message should contain specific details about the problem, e.g.:

    - Request field x.y.z is xxx, expected one of [yyy, zzz].
    - Parameter 'age' is out of range [0, 125].

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_400_BAD_REQUEST
    _DEFAULT_MESSAGE = "An invalid argument was specified by the client. "

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the error.

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """
        super(ClientValueError, self).__init__(
            status_code=ClientValueError._HTTP_STATUS_CODE,
            message=message or ClientValueError._DEFAULT_MESSAGE,
            explanation=explanation,
            metadata=metadata,
            resource=resource,
        )


class ResourceUpdateFailedError(ClientBaseError):
    """Client error that indicates that a requested update for a resource failed.

    The error message should contain specific details about the problem and resource, e.g.:

    - Unable to apply patch update for resource 'xxx'.

    The error details will be shown to the client (user) if it is not handled otherwise.
    """

    _HTTP_STATUS_CODE = status.HTTP_409_CONFLICT
    _DEFAULT_MESSAGE = "Resource update failed. Try again."
    _DEFAULT_EXPLANATION = "The requested resource updated could not be applied. The reason could concurrent or conflicting modifications to the resource. Please try the update again."

    def __init__(
        self,
        message: Optional[str] = None,
        explanation: Optional[str] = None,
        metadata: Optional[Dict] = None,
        resource: Optional[str] = None,
    ) -> None:
        """Initializes the error.

        Args:
            message (optional): A message shown to the user that overwrites the default message.
            explanation (optional): A human readable explanation specific to this error that is helpful to locate the problem and give advice on how to proceed.
            metadata (optional): Additional problem details/metadata.
            resource (optional): A resource name (relative URI reference) of a specific resource instance associated with the error.
        """
        super(ResourceUpdateFailedError, self).__init__(
            status_code=ResourceUpdateFailedError._HTTP_STATUS_CODE,
            message=message or ResourceUpdateFailedError._DEFAULT_MESSAGE,
            explanation=explanation or ResourceUpdateFailedError._DEFAULT_EXPLANATION,
            metadata=metadata,
            resource=resource,
        )


CREATE_RESOURCE_RESPONSES: Mapping[Union[int, str], Dict[str, Any]] = {
    status.HTTP_409_CONFLICT: {
        "description": "The resource already exists.",
        "model": ProblemDetails,
    }
}

GET_RESOURCE_RESPONSES: Mapping[Union[int, str], Dict[str, Any]] = {
    status.HTTP_404_NOT_FOUND: {
        "description": "A specified resource is not found.",
        "model": ProblemDetails,
    }
}

UPDATE_RESOURCE_RESPONSES: Mapping[Union[int, str], Dict[str, Any]] = {
    **GET_RESOURCE_RESPONSES,
    status.HTTP_409_CONFLICT: {
        "description": "Resource update failed.",
        "model": ProblemDetails,
    },
}

AUTH_ERROR_RESPONSES: Mapping[Union[int, str], Dict[str, Any]] = {
    status.HTTP_401_UNAUTHORIZED: {
        "description": "Invalid authentication credentials.",
        "model": ProblemDetails,
    },
    status.HTTP_403_FORBIDDEN: {
        "description": "The authorized user does not have sufficient permission.",
        "model": ProblemDetails,
    },
}

VALIDATION_ERROR_RESPONSE: Mapping[Union[int, str], Dict[str, Any]] = {
    status.HTTP_422_UNPROCESSABLE_ENTITY: {
        "description": "Validation Error.",
        "model": ProblemDetails,
    },
}



from typing import Dict, Iterator, List, Optional

import requests
from pydantic import parse_obj_as

from contaxy.clients.shared import handle_errors
from contaxy.operations.file import FileOperations
from contaxy.schema import File, FileInput, FileStream, ResourceAction


class FileClient(FileOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    def list_files(
        self,
        project_id: str,
        recursive: bool = True,
        include_versions: bool = False,
        prefix: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> List[File]:

        query_params: Dict = {
            "recursive": recursive,
            "include_versions": include_versions,
        }
        if prefix:
            query_params.update({"prefix": prefix})

        response = self._client.get(
            f"/projects/{project_id}/files",
            params=query_params,
            **request_kwargs,
        )
        handle_errors(response)
        return parse_obj_as(List[File], response.json())

    def get_file_metadata(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> File:

        query_params: Dict = {}
        if version:
            query_params.update({"version": version})

        response = self._client.get(
            f"/projects/{project_id}/files/{file_key}:metadata",
            params=query_params,
            **request_kwargs,
        )
        handle_errors(response)
        return parse_obj_as(File, response.json())

    def update_file_metadata(
        self,
        file: FileInput,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> File:
        query_params: Dict = {}
        if version:
            query_params.update({"version": version})

        response = self._client.patch(
            f"/projects/{project_id}/files/{file_key}",
            params=query_params,
            data=file.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(response)
        return parse_obj_as(File, response.json())

    def upload_file(
        self,
        project_id: str,
        file_key: str,
        file_stream: FileStream,
        content_type: str = "application/octet-stream",
        request_kwargs: Dict = {},
    ) -> File:
        # ! It is strongly recommended that you open files in binary mode. This is because Requests may attempt to provide the Content-Length header for you, and if it does this value will be set to the number of bytes in the file. Errors may occur if you open the file in text mode.

        # TODO Check if it actually gets streamed or not
        # ! In the event you are posting a very large file as a multipart/form-data request, you may want to stream the request. By default, requests does not support this, but there is a separate package which does - requests-toolbelt. You should read the toolbelt’s documentation for more details about how to use it.

        response = self._client.post(
            f"/projects/{project_id}/files/{file_key}",
            files={"file": (f"{file_key}", file_stream, content_type)},
            **request_kwargs,
        )
        handle_errors(response)
        return parse_obj_as(File, response.json())

    def download_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> Iterator[bytes]:
        query_params: Dict = {}
        if version:
            query_params.update({"version": version})

        response = self._client.get(
            f"/projects/{project_id}/files/{file_key}:download",
            params=query_params,
            stream=True,
            **request_kwargs,
        )
        handle_errors(response)
        return response.iter_content(chunk_size=10 * 1024 * 1024)

    def delete_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        keep_latest_version: bool = False,
        request_kwargs: Dict = {},
    ) -> None:
        query_params: Dict = {"keep_latest_version": keep_latest_version}
        if version:
            query_params.update({"version": version})

        response = self._client.delete(
            f"/projects/{project_id}/files/{file_key}",
            params=query_params,
            **request_kwargs,
        )
        handle_errors(response)

    def delete_files(
        self,
        project_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self._client.delete(
            f"/projects/{project_id}/files",
            **request_kwargs,
        )
        handle_errors(response)

    def list_file_actions(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> List[ResourceAction]:
        # TODO
        pass

    def execute_file_action(
        self,
        project_id: str,
        file_key: str,
        action_id: str,
        version: Optional[str] = None,
        request_kwargs: Dict = {},
    ) -> None:
        # TODO
        pass



from typing import Any, List, Optional

from fastapi import APIRouter, Depends, Query, Response, status
from fastapi.param_functions import Body

from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
    get_optional_api_token,
)
from contaxy.schema import CoreOperations, User, UserInput, UserRegistration
from contaxy.schema.auth import USER_ID_PARAM, AccessLevel, TokenPurpose, TokenType
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    UPDATE_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
    ClientBaseError,
    PermissionDeniedError,
)
from contaxy.schema.system import SystemState
from contaxy.utils import auth_utils, id_utils

router = APIRouter(
    tags=["users"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)


@router.get(
    "/users",
    operation_id=CoreOperations.LIST_USERS.value,
    response_model=List[User],
    response_model_exclude_unset=True,
    summary="List all users.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
)
def list_users(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all users that are visible to the authenticated user."""
    component_manager.verify_access(
        token, "users", AccessLevel.READ
    )  # TODO: the right permission?
    return component_manager.get_auth_manager().list_users()


@router.post(
    "/users",
    operation_id=CoreOperations.CREATE_USER.value,
    response_model=User,
    response_model_exclude_unset=True,
    summary="Create a user.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def create_user(
    user_input: UserRegistration,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_optional_api_token),
) -> Any:
    """Creates a user. For the user also a technical project is created.

    If the `password` is not provided, the user can only login by using other methods (social login).
    """
    system_info = component_manager.get_system_manager().get_system_info()
    if system_info.system_state != SystemState.RUNNING:
        raise ClientBaseError(
            status.HTTP_400_BAD_REQUEST,
            "User registration is not possible before system has been initialized!",
        )

    if not component_manager.global_state.settings.USER_REGISTRATION_ENABLED:
        if token is not None:
            # An admin can create users even if registration is disabled
            component_manager.verify_access(token, "users", AccessLevel.ADMIN)
        else:
            raise PermissionDeniedError(
                "User self-registration is deactivated. Please contact an administrator."
            )

    # If self registration is enabled, everyone can create users
    user = auth_utils.create_and_setup_user(
        user_input=user_input,
        auth_manager=component_manager.get_auth_manager(),
        project_manager=component_manager.get_project_manager(),
    )

    # TODO: return also user_project?
    return user


@router.get(
    "/users/me",
    operation_id=CoreOperations.GET_MY_USER.value,
    response_model=User,
    response_model_exclude_unset=True,
    summary="Get my user metadata.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_my_user(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the user metadata from the authenticated user."""
    authorized_access = component_manager.verify_access(token)

    # Check if read access to user object is allowed
    component_manager.verify_access(
        token, authorized_access.authorized_subject, AccessLevel.READ
    )

    return component_manager.get_auth_manager().get_user(
        id_utils.extract_user_id_from_resource_name(
            authorized_access.authorized_subject
        )
    )


@router.get(
    "/users/{user_id}",
    operation_id=CoreOperations.GET_USER.value,
    response_model=User,
    response_model_exclude_unset=True,
    summary="Get user metadata.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_user(
    user_id: str = USER_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the user metadata for a single user."""
    component_manager.verify_access(token, f"users/{user_id}", AccessLevel.READ)
    return component_manager.get_auth_manager().get_user(user_id)


@router.patch(
    "/users/{user_id}",
    operation_id=CoreOperations.UPDATE_USER.value,
    response_model=User,
    response_model_exclude_unset=True,
    summary="Update user metadata.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
    responses={**UPDATE_RESOURCE_RESPONSES},
)
def update_user(
    user_id: str = USER_ID_PARAM,
    user_input: UserInput = Body(...),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Updates the user metadata.

    This will update only the properties that are explicitly set in the patch request.
    The patching is based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).
    """
    # TODO: should only be called by an admin with access to all users, not by users itself
    component_manager.verify_access(token, "users", AccessLevel.ADMIN)
    return component_manager.get_auth_manager().update_user(user_id, user_input)


@router.put(
    "/users/{user_id}:change-password",
    operation_id=CoreOperations.CHANGE_PASSWORD.value,
    summary="Change the user password",
    tags=["users"],
    status_code=status.HTTP_204_NO_CONTENT,
    responses={**GET_RESOURCE_RESPONSES},
)
def change_password(
    user_id: str = USER_ID_PARAM,
    password: str = Body(...),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Changes the password of a given user.

    The password can be changed by the given user or a system admin.

    The password is stored as a hash.
    """
    # TODO: check existing password as well for normal users
    if not component_manager.global_state.settings.PASSWORD_AUTH_ENABLED:
        # Admins still can set and change password
        component_manager.verify_access(token, "users", AccessLevel.ADMIN)
    else:
        # Only check if token allows admin access on user object
        component_manager.verify_access(token, f"users/{user_id}", AccessLevel.ADMIN)

    component_manager.get_auth_manager().change_password(user_id, password)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/users/{user_id}",
    operation_id=CoreOperations.DELETE_USER.value,
    summary="Delete a user.",
    tags=["users"],
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_user(
    user_id: str = USER_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes a user and all resources which are only accesible by this user.

    Shared project resources will not be deleted.
    """
    component_manager.verify_access(token, "users/{user_id}", AccessLevel.ADMIN)
    component_manager.get_auth_manager().delete_user(user_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/users/{user_id}/token",
    operation_id=CoreOperations.GET_USER_TOKEN.value,
    response_model=str,
    summary="Get a user token.",
    tags=["users"],
    status_code=status.HTTP_200_OK,
)
def get_user_token(
    user_id: str = USER_ID_PARAM,
    access_level: AccessLevel = Query(
        AccessLevel.WRITE,
        description="Access level of the token.",
        type="string",
    ),
    description: Optional[str] = Query(
        None, description="Attach a short description to the generated token."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns an API token with permission to access all resources accesible by the given user.

    The `read` access level allows read-only access on all resources.
    The `write` access level allows to create and delete user resources.
    The `admin` access level allows additional user actions such as deletion of the user itself.
    """
    # Only allow creating tokens if user has admin access to the user object
    authorized_access = component_manager.verify_access(
        token, f"users/{user_id}", AccessLevel.ADMIN
    )

    # Provide access to all resources from the user
    user_token_scope = auth_utils.construct_permission("*", access_level)

    return component_manager.get_auth_manager().create_token(
        scopes=[user_token_scope],
        token_type=TokenType.API_TOKEN,
        token_subject=authorized_access.authorized_subject,
        token_purpose=TokenPurpose.USER_API_TOKEN,
        description=description,
    )



import asyncio
import hashlib
from collections.abc import AsyncGenerator, Generator
from hashlib import sha1
from typing import IO, Any, Callable, Mapping, Optional

import filetype
from streaming_form_data import StreamingFormDataParser
from streaming_form_data.targets import BaseTarget

from contaxy.schema.file import FileStream


class MultipartStreamTarget(BaseTarget):
    """StreamTarget stores one chunk at a time in-memory and deletes it upon read.

    This is useful in case you'd like to have a parsed stream.
    """

    def __init__(
        self, validator: Optional[Callable] = None, hash_algo: Optional[str] = "md5"
    ) -> None:
        super().__init__(validator)

        self._chunk = b""
        self._hash = hashlib.new(hash_algo) if hash_algo else None

    def on_data_received(self, chunk: bytes) -> None:
        self._chunk = chunk
        if self._hash:
            self._hash.update(self._chunk)

    @property
    def value(self) -> bytes:
        chunk = self._chunk
        self._chunk = b""
        return chunk

    @property
    def hash(self) -> str:
        # Todo: Throw error
        if not self._hash:
            raise ValueError("No hash algorithm was set during object intialization")
        return self._hash.hexdigest()


class FormMultipartStream(FileStream):
    def __init__(
        self,
        stream: Generator,
        headers: Mapping[str, str],
        form_field: str = "file",
        pre_read_size: int = 1024,
        hash_algo: Optional[str] = "md5",
    ) -> None:
        """Currently, it is a requirement that the multipart/form-data stream embodies only one form field, i.e. the form_field parameter.

        Args:
            stream (Generator): The multipart/form-data byte stream.
            headers (Mapping[str, str]): Http header of the reuqest.
            form_field (str, optional): The name of the form field used to upload the file. Defaults to "file".
            pre_read_size (int): The pre_read_size controls the byte size which will be used to determine some metadata in advance e.g. a filename based on the content-disposition. Defaults to 1024.
            hash_algo (Optional[str]): Name of the hash algorithm supported by python's hashlib. If not set, now file hash will be calculated.
        """
        self._stream = stream
        self._parser = StreamingFormDataParser(headers=headers)
        self._stream_target = MultipartStreamTarget(hash_algo=hash_algo)
        self._parser.register(form_field, self._stream_target)
        self.is_stream_finished = False
        self._unprocessed_bytes = b""
        # This will try to parse the metadata from the beginning of the stream
        self._unprocessed_bytes = self.read(pre_read_size) + self._unprocessed_bytes
        # Try guess mime type based on magic numbers
        self._guessed_content_type = filetype.guess_mime(self._unprocessed_bytes)

    @property
    def filename(self) -> Optional[str]:
        return self._stream_target.multipart_filename

    @property
    def content_type(self) -> Optional[str]:
        default_mime = "application/octet-stream"
        if (
            self._stream_target.multipart_content_type
            and self._stream_target.multipart_content_type != default_mime
        ):
            return self._stream_target.multipart_content_type
        elif self._guessed_content_type:
            return self._guessed_content_type
        else:
            return default_mime

    @property
    def hash(self) -> str:
        try:
            return self._stream_target.hash
        except ValueError:
            return ""

    def read(self, size: int = -1) -> bytes:

        if size == 0:
            return b""

        unprocessed_bytes_length = len(self._unprocessed_bytes)

        while (
            unprocessed_bytes_length < size or size == -1
        ) and not self.is_stream_finished:
            try:
                raw_chunk = self._stream.__next__()
            except StopIteration:
                self.is_stream_finished = True
                break

            self._parser.data_received(raw_chunk)
            parsed_chunk = self._stream_target.value
            self._unprocessed_bytes += parsed_chunk
            unprocessed_bytes_length = len(self._unprocessed_bytes)

        if unprocessed_bytes_length > size and size > -1:
            required_bytes = max(size - unprocessed_bytes_length, size)
            bytes_received = self._unprocessed_bytes[:required_bytes]
            self._unprocessed_bytes = self._unprocessed_bytes[required_bytes:]
        else:
            bytes_received = self._unprocessed_bytes
            self._unprocessed_bytes = b""

        return bytes_received


class SyncFromAsyncGenerator(Generator):
    """This genrator implementation wraps an async generator to make it compatible for sync implementations."""

    def __init__(
        self,
        async_generator: AsyncGenerator,
        event_loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__()
        self._generator = async_generator
        self._event_loop = event_loop

    def __next__(self) -> Any:
        try:
            data = asyncio.run_coroutine_threadsafe(
                self._generator.__anext__(), self._event_loop
            ).result()
            return data
        except StopAsyncIteration:
            raise StopIteration()

    def send(self, value) -> Any:  # type: ignore
        data = asyncio.run_coroutine_threadsafe(
            self._generator.asend(value), self._event_loop
        ).result()
        return data

    def throw(self, typ, val=None, tb=None) -> None:  # type: ignore
        asyncio.run_coroutine_threadsafe(
            self._generator.athrow(typ, val, tb), self._event_loop
        )

    def close(self) -> None:
        asyncio.run_coroutine_threadsafe(self._generator.aclose(), self._event_loop)

    def __iter__(self) -> Generator:
        return self


def generate_file_id(file_name: str, version_id: str) -> str:
    return sha1(bytes(f"{file_name}{version_id}", "utf8")).hexdigest()


class FileStreamWrapper(FileStream):
    def __init__(self, stream: IO[bytes], hash_algo: str = "md5"):
        self.stream = stream
        self._hash = hashlib.new(hash_algo)

    @property
    def hash(self) -> str:
        return self._hash.hexdigest()

    def read(self, size: int = -1) -> bytes:
        chunk = self.stream.read(size)
        self._hash.update(chunk)
        return chunk



import asyncio
import functools
from typing import Any, Dict

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from fastapi.openapi.utils import get_openapi
from loguru import logger
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import RedirectResponse

from contaxy import __version__, config
from contaxy.api import error_handling
from contaxy.api.endpoints import (
    auth,
    deployment,
    extension,
    file,
    json_db,
    project,
    seed,
    system,
    user,
)
from contaxy.managers.components import ComponentManager
from contaxy.managers.deployment.utils import stop_idle_services
from contaxy.utils import fastapi_utils, state_utils

# Initialize API
app = FastAPI(
    title="Contaxy API",
    description="Functionality to create and manage projects, services, jobs, and files.",
    version=__version__,
)

if config.settings.DEBUG:
    fastapi_utils.add_timing_info(app)


# Custom Exception Handling
@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(request, exc):  # type: ignore
    return await error_handling.handel_http_exeptions(request, exc)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):  # type: ignore
    return await error_handling.handle_validation_exception(request, exc)


# Startup and shutdown events
@app.on_event("startup")
def on_startup() -> None:
    """Initializes the global app state object."""
    logger.info("Starting API server instance.")
    state_utils.GlobalState(app.state).settings = config.settings
    state_utils.GlobalState(
        app.state
    ).shared_namespace.async_loop = asyncio.get_running_loop()
    component_manager = ComponentManager.from_app(app)
    # Schedule regular cleanup of idle services
    fastapi_utils.schedule_call(
        func=functools.partial(stop_idle_services, component_manager),
        interval=config.settings.SERVICE_IDLE_CHECK_INTERVAL,
    )


@app.on_event("shutdown")
def on_shutdown() -> None:
    """Closes the global app state object.

    This also calls all registered close callback functions.
    """
    logger.info("Stopping API server instance.")
    state_utils.GlobalState(app.state).close()


if config.settings.BACKEND_CORS_ORIGINS:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=[str(origin) for origin in config.settings.BACKEND_CORS_ORIGINS],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


# Redirect to docs
@app.get("/", include_in_schema=False)
def root() -> Any:
    return RedirectResponse("./docs")


app.include_router(system.router)
app.include_router(auth.router)
app.include_router(user.router)
app.include_router(project.router)
app.include_router(deployment.job_router)
app.include_router(deployment.service_router)
app.include_router(extension.router)
app.include_router(file.router)
app.include_router(json_db.router)

if config.settings.DEBUG:
    app.include_router(seed.router)


def custom_openapi() -> Dict[str, Any]:
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title=app.title,
        description=app.description,
        version=app.version,
        routes=app.routes,
    )

    app.openapi_schema = file.modify_openapi_schema(openapi_schema)
    return app.openapi_schema


app.openapi = custom_openapi  # type: ignore

# Patch Fastapi to allow relative path resolution.
fastapi_utils.patch_fastapi(app)
